from __future__ import annotations

import random


def test_consistent():
    from dials.array_family import flex
    from dials.model.data import Shoebox

    shoebox = flex.shoebox(10)

    for i in range(10):
        x0 = random.randint(0, 1000)
        y0 = random.randint(0, 1000)
        z0 = random.randint(0, 1000)
        x1 = random.randint(1, 10) + x0
        y1 = random.randint(1, 10) + y0
        z1 = random.randint(1, 10) + z0
        shoebox[i] = Shoebox((x0, x1, y0, y1, z0, z1))

    assert shoebox.is_consistent() == flex.bool(10, False)

    for i in range(10):
        shoebox[i].allocate()

    assert shoebox.is_consistent() == flex.bool(10, True)

    for i in [0, 2, 4, 6, 8]:
        shoebox[i].data.resize(flex.grid(10, 10, 10))

    assert shoebox.is_consistent() == flex.bool([False, True] * 5)

    for i in range(10):
        shoebox[i].deallocate()

    assert shoebox.is_consistent() == flex.bool(10, False)


def test_is_bbox_within_image_volume():
    from dials.array_family import flex
    from dials.model.data import Shoebox

    isize = (1000, 1000)
    srange = (0, 100)

    shoebox = flex.shoebox(7)
    shoebox[0] = Shoebox((10, 20, 10, 20, 10, 20))
    shoebox[1] = Shoebox((-10, 20, 10, 20, 10, 20))
    shoebox[2] = Shoebox((10, 20, -10, 20, 10, 20))
    shoebox[3] = Shoebox((10, 20, 10, 20, -10, 20))
    shoebox[4] = Shoebox((10, 1020, 10, 20, 10, 20))
    shoebox[5] = Shoebox((10, 20, 10, 1020, 10, 20))
    shoebox[6] = Shoebox((10, 20, 10, 20, 10, 1020))

    assert shoebox.is_bbox_within_image_volume(isize, srange) == flex.bool(
        [True, False, False, False, False, False, False]
    )


def test_does_bbox_contain_bad_pixels():
    from dials.array_family import flex
    from dials.model.data import Shoebox

    mask = flex.bool(flex.grid(100, 100), True)
    for j in range(100):
        for i in range(40, 60):
            mask[j, i] = False
            mask[i, j] = False

    shoebox = flex.shoebox(1000)
    res = flex.bool(1000)
    for i in range(1000):
        x0 = random.randint(0, 90)
        y0 = random.randint(0, 90)
        z0 = random.randint(0, 90)
        x1 = random.randint(1, 10) + x0
        y1 = random.randint(1, 10) + y0
        z1 = random.randint(1, 10) + z0

        shoebox[i] = Shoebox((x0, x1, y0, y1, z0, z1))

        res2 = False
        if x0 >= 40 and x0 < 60:
            res2 = True
        if x1 > 40 and x1 <= 60:
            res2 = True
        if y0 >= 40 and y0 < 60:
            res2 = True
        if y1 > 40 and y1 <= 60:
            res2 = True

        res[i] = res2

    assert shoebox.does_bbox_contain_bad_pixels(mask) == res


def test_count_mask_values():
    from dials.array_family import flex
    from dials.model.data import Shoebox

    shoebox = flex.shoebox(10)
    num = flex.int(10)
    value = 1 << 2
    for i in range(10):
        x0 = random.randint(0, 90)
        y0 = random.randint(0, 90)
        z0 = random.randint(0, 90)
        x1 = random.randint(1, 10) + x0
        y1 = random.randint(1, 10) + y0
        z1 = random.randint(1, 10) + z0

        shoebox[i] = Shoebox((x0, x1, y0, y1, z0, z1))
        shoebox[i].allocate()
        maxnum = len(shoebox[i].mask)
        num[i] = random.randint(1, maxnum)
        indices = random.sample(list(range(maxnum)), num[i])
        for j in indices:
            shoebox[i].mask[j] = value

    assert shoebox.count_mask_values(value) == num


def test_bounding_boxes():
    from dials.array_family import flex
    from dials.model.data import Shoebox

    shoebox = flex.shoebox(10)
    bbox = flex.int6(10)
    for i in range(10):
        x0 = random.randint(0, 90)
        y0 = random.randint(0, 90)
        z0 = random.randint(0, 90)
        x1 = random.randint(1, 10) + x0
        y1 = random.randint(1, 10) + y0
        z1 = random.randint(1, 10) + z0
        bbox[i] = (x0, x1, y0, y1, z0, z1)
        shoebox[i] = Shoebox(bbox[i])

    bbox2 = shoebox.bounding_boxes()
    for i in range(10):
        assert bbox2[i] == bbox[i]



"""Definitions of screw axes and space groups for Laue classes."""

from __future__ import annotations

from dials.algorithms.symmetry.absences.screw_axes import (
    ScrewAxis21a,
    ScrewAxis21b,
    ScrewAxis21c,
    ScrewAxis31c,
    ScrewAxis41a,
    ScrewAxis41b,
    ScrewAxis41c,
    ScrewAxis42c,
    ScrewAxis61c,
    ScrewAxis62c,
    ScrewAxis63c,
    ScrewAxisObserver,
)


def score_screw_axes(
    laue_group_info, reflection_table, significance_level=0.95, method="direct"
):
    """Get the relevant screw axes and score them. Print pretty table."""
    scores = []
    axes = []
    for axis in laue_group_info["unique_axes"]:
        a = axis()
        a.register_observer(event="scored axis", observer=ScrewAxisObserver())
        if "equivalent_axes" in laue_group_info:
            if axis in laue_group_info["equivalent_axes"]:
                for equivalent in laue_group_info["equivalent_axes"][axis]:
                    a.add_equivalent_axis(equivalent())
        scores.append(a.score_axis(reflection_table, significance_level, method))
        axes.append(a)
    return axes, scores


def score_space_groups(screw_axis_scores, laue_group_info):
    """Score the space groups in a laue group based on axes scores."""
    space_groups = []
    scores = []
    for space_g, conditions in laue_group_info["space_groups"]:
        sg_score = 1.0
        for condition, score in zip(conditions, screw_axis_scores):
            # condition can be True, False or 'pass'
            if condition is True:
                sg_score *= score
            elif condition is False:
                sg_score *= 1.0 - score
        scores.append(sg_score)
        space_groups.append(space_g)
    return space_groups, scores


### For each laue group, define allowed screw axes and how to test these to
### determine the space group; True, False or 'pass' (if not relevant).
### If relevant, list equivalent axes and enantiomorphic pairs.
p2m = {
    "unique_axes": [ScrewAxis21b],
    "space_groups": [("P 2", [False]), ("P 21", [True])],
}

Cmmm = {
    "unique_axes": [ScrewAxis21c],
    "space_groups": [
        ("C 2 2 2", [False]),
        ("C 2 2 21", [True]),  # Assuming we don't allow B 2 21 2, A 21 2 2
    ],
}

pmmm = {
    "unique_axes": [ScrewAxis21a, ScrewAxis21b, ScrewAxis21c],
    "space_groups": [
        ("P 2 2 2", [False, False, False]),
        ("P 2 2 21", [False, False, True]),
        ("P 2 21 2", [False, True, False]),
        ("P 21 2 2", [True, False, False]),
        # ^ three equivalent groups, but keep separate to allow standard setting.
        ("P 21 21 2", [True, True, False]),
        ("P 21 2 21", [True, False, True]),
        ("P 2 21 21", [False, True, True]),
        ("P 21 21 21", [True, True, True]),
        # ^ three equivalent groups, but keep separate to allow standard setting.
    ],
}

p6m = {
    "unique_axes": [ScrewAxis61c, ScrewAxis62c, ScrewAxis63c],
    "space_groups": [
        ("P 6", [False, False, False]),
        ("P 61", [True]),
        ("P 62", [False, True]),
        ("P 63", [False, "pass", True]),
    ],
    "enantiomorphic pairs": {"P 61": "P 65", "P 62": "P 64"},
}

p6mmm = {
    "unique_axes": [ScrewAxis61c, ScrewAxis62c, ScrewAxis63c],
    "space_groups": [
        ("P 6 2 2", [False, False, False]),
        ("P 61 2 2", [True]),
        ("P 62 2 2", [False, True]),
        ("P 63 2 2", [False, "pass", True]),
    ],
    "enantiomorphic pairs": {"P 61 2 2": "P 65 2 2", "P 62 2 2": "P 64 2 2"},
}

p3 = {
    "unique_axes": [ScrewAxis31c],
    "space_groups": [("P 3", [False]), ("P 31", [True])],
    "enantiomorphic pairs": {"P 31": "P 32"},
}

p31m = {
    "unique_axes": [ScrewAxis31c],
    "space_groups": [("P 3 1 2", [False]), ("P 31 1 2", [True])],
    "enantiomorphic pairs": {"P 31 1 2": "P 32 1 2"},
}

p3m1 = {
    "unique_axes": [ScrewAxis31c],
    "space_groups": [("P 3 2 1", [False]), ("P 31 2 1", [True])],
    "enantiomorphic pairs": {"P 31 2 1": "P 32 2 1"},
}

p4mmm = {
    "unique_axes": [ScrewAxis41c, ScrewAxis21a, ScrewAxis42c],
    "equivalent_axes": {ScrewAxis21a: [ScrewAxis21b]},
    "space_groups": [
        ("P 4 2 2", [False, False, False]),
        ("P 4 21 2", [False, True, False]),
        ("P 41 2 2", [True, False]),
        ("P 42 2 2", [False, False, True]),
        ("P 41 21 2", [True, True]),
        ("P 42 21 2", [False, True, True]),
    ],
    "enantiomorphic pairs": {"P 41 2 2": "P 43 2 2", "P 41 21 2": "P 43 21 2"},
}

p4m = {
    "unique_axes": [ScrewAxis41c, ScrewAxis42c],
    "space_groups": [
        ("P 4", [False, False]),
        ("P 41", [True]),
        ("P 42", [False, True]),
    ],
    "enantiomorphic pairs": {"P 41": "P 43"},
}

I4m = {
    "unique_axes": [ScrewAxis41c],
    "space_groups": [("I 4", [False]), ("I 41", [True])],
}

I4mmm = {
    "unique_axes": [ScrewAxis41c],
    "space_groups": [("I 4 2 2", [False]), ("I 41 2 2", [True])],
}

pm3 = {
    "unique_axes": [ScrewAxis21a],
    "equivalent_axes": {ScrewAxis21a: [ScrewAxis21b, ScrewAxis21c]},
    "space_groups": [("P 2 3", [False]), ("P 21 3", [True])],
}

pm3m = {
    "unique_axes": [ScrewAxis41a, ScrewAxis42c],
    "equivalent_axes": {ScrewAxis41a: [ScrewAxis41b, ScrewAxis41c]},
    "space_groups": [
        ("P 4 3 2", [False, False]),
        ("P 41 3 2", [True]),
        ("P 42 3 2", [False, True]),
    ],
    "enantiomorphic pairs": {"P 41 3 2": "P 43 3 2"},
}

Im3m = {
    "unique_axes": [ScrewAxis41a],
    "equivalent_axes": {ScrewAxis41a: [ScrewAxis41b, ScrewAxis41c]},
    "space_groups": [("I 4 3 2", [False]), ("I 41 3 2", [True])],
}

Fm3m = {
    "unique_axes": [ScrewAxis41a],
    "equivalent_axes": {ScrewAxis41a: [ScrewAxis41b, ScrewAxis41c]},
    "space_groups": [("F 4 3 2", [False]), ("F 41 3 2", [True])],
}

# These laue groups contain 59 space groups (discarding the equivalent defined in P222)
# - adding P1, C2, F222, R3, R32 and F23 make 65 MX space groups.
laue_groups = {
    "P 1 2/m 1": p2m,
    "P m m m": pmmm,
    "C m m m": Cmmm,
    # "I m m m": Immm, #space groups are special pairs due to lattice centering
    "P 4/m": p4m,
    "I 4/m": I4m,
    "P 4/m m m": p4mmm,
    "I 4/m m m": I4mmm,
    "P 6/m": p6m,
    "P 6/m m m": p6mmm,
    "P -3": p3,
    "P -3 1 m": p31m,
    "P -3 m 1": p3m1,
    "P m -3": pm3,
    # "I m -3": Im3, #space groups are special pairs due to lattice centering
    "P m -3 m": pm3m,
    "I m -3 m": Im3m,
    "F m -3 m": Fm3m,
}

# special pairs due to centering - see either
# https://strucbio.biologie.uni-konstanz.de/xdswiki/index.php/Space_group_determination
# International Tables for Crystallography Vol. A sec. 3.1.5



from __future__ import annotations

import pytest

from dxtbx.serialize import load

from dials.algorithms.spot_finding import per_image_analysis
from dials.array_family import flex


@pytest.fixture
def centroid_test_data(dials_data):
    reflections = flex.reflection_table.from_file(
        dials_data("centroid_test_data", pathlib=True) / "strong.pickle"
    )
    experiments = load.experiment_list(
        dials_data("centroid_test_data", pathlib=True) / "imported_experiments.json"
    )
    reflections.centroid_px_to_mm(experiments)
    reflections.map_centroids_to_reciprocal_space(experiments)
    return experiments, reflections


def test_estimate_resolution_limit_distl_method1(centroid_test_data, tmp_path):
    experiments, reflections = centroid_test_data
    plot_file = tmp_path / "distl1.png"
    d_min, noisiness = per_image_analysis.estimate_resolution_limit_distl_method1(
        reflections=reflections, plot_filename=plot_file
    )
    assert d_min == pytest.approx(1.8771983880778702)
    assert noisiness == pytest.approx(0.021021021021021023)
    assert plot_file.is_file()


def test_estimate_resolution_limit_distl_method2(centroid_test_data, tmp_path):
    experiments, reflections = centroid_test_data
    plot_file = tmp_path / "distl2.png"
    d_min, noisiness = per_image_analysis.estimate_resolution_limit_distl_method2(
        reflections=reflections, plot_filename=plot_file
    )
    assert d_min == pytest.approx(1.6293601446185495)
    assert noisiness == pytest.approx(0.0858974358974359)
    assert plot_file.is_file()


def test_estimate_resolution(centroid_test_data, tmp_path):
    experiments, reflections = centroid_test_data
    ice_sel = per_image_analysis.ice_rings_selection(reflections, width=0.004)
    assert ice_sel.count(True) == 76
    plot_file = tmp_path / "i_over_sigi_vs_resolution.png"
    d_min = per_image_analysis.estimate_resolution_limit(
        reflections=reflections, plot_filename=plot_file
    )
    assert d_min == pytest.approx(1.446715534174674)
    assert plot_file.is_file()
    d_min = per_image_analysis.estimate_resolution_limit(
        reflections=reflections, ice_sel=ice_sel
    )
    assert d_min == pytest.approx(1.446715534174674)


def test_stats_for_reflection_table(centroid_test_data):
    _, reflections = centroid_test_data
    stats = per_image_analysis.stats_for_reflection_table(reflections)
    result = stats._asdict()
    expected = {
        "d_min_distl_method_1": 1.8771983880778702,
        "d_min_distl_method_2": 1.6293601446185495,
        "estimated_d_min": 1.446715534174674,
        "n_spots_4A": 76,
        "n_spots_no_ice": 578,
        "n_spots_total": 654,
        "noisiness_method_1": 0.021021021021021023,
        "noisiness_method_2": 0.0858974358974359,
        "total_intensity": 919847.0,
    }
    for k, v in expected.items():
        assert result[k] == pytest.approx(v)


def test_stats_for_reflection_table_no_resolution_analysis_no_ice_filtering(
    centroid_test_data,
):
    _, reflections = centroid_test_data
    stats = per_image_analysis.stats_for_reflection_table(
        reflections, resolution_analysis=False, filter_ice=False
    )
    result = stats._asdict()
    expected = {
        "d_min_distl_method_1": -1.0,
        "d_min_distl_method_2": -1.0,
        "estimated_d_min": -1.0,
        "n_spots_4A": 76,
        "n_spots_no_ice": 654,
        "n_spots_total": 654,
        "noisiness_method_1": -1.0,
        "noisiness_method_2": -1.0,
        "total_intensity": 961472.0,
    }
    for k, v in expected.items():
        assert result[k] == pytest.approx(v)


def test_stats_per_image(centroid_test_data):
    experiments, reflections = centroid_test_data
    stats = per_image_analysis.stats_per_image(experiments[0], reflections)
    result = stats._asdict()
    for v in result.values():
        assert len(v) == len(experiments[0].scan)
    assert stats.n_spots_total == [90, 100, 67, 49, 54, 62, 68, 83, 81]
    t = stats.as_table()
    assert len(t) == len(experiments[0].scan) + 1
    assert t[0] == [
        "image",
        "#spots",
        "#spots_no_ice",
        "total_intensity",
        "d_min",
        "d_min (distl method 1)",
        "d_min (distl method 2)",
    ]
    assert t[1] == ["1", "90", "77", "28214", "1.56", "2.08 (0.09)", "1.59 (0.27)"]
    # Test n_rows option
    t = stats.as_table(n_rows=3)
    assert len(t) == 4
    # Test perm option
    perm = flex.random_permutation(len(experiments[0].scan))
    t = stats.as_table(perm=perm)
    assert [tt[0] for tt in t[1:]] == [str(i + 1) for i in perm]


def test_stats_table_no_resolution_analysis(centroid_test_data):
    experiments, reflections = centroid_test_data
    stats = per_image_analysis.stats_per_image(
        experiments[0], reflections, resolution_analysis=False
    )
    t = stats.as_table()
    assert t[0] == ["image", "#spots", "#spots_no_ice", "total_intensity"]


def test_plot_stats(centroid_test_data, tmp_path):
    experiments, reflections = centroid_test_data
    stats = per_image_analysis.stats_per_image(
        experiments[0], reflections, resolution_analysis=False
    )
    image_file = tmp_path / "pia.png"
    per_image_analysis.plot_stats(stats, filename=image_file)
    assert image_file.is_file()



from __future__ import annotations

import copy
import logging
import math
import operator

import libtbx.phil
from cctbx import miller
from dxtbx.model import Crystal
from scitbx import matrix
from scitbx.math import least_squares_plane, superpose

from dials.algorithms.indexing import DialsIndexError
from dials.array_family import flex

from .strategy import Strategy

logger = logging.getLogger(__name__)

TWO_PI = 2.0 * math.pi
FIVE_DEG = TWO_PI * 5.0 / 360.0


class CompleteGraph:
    def __init__(self, seed_vertex):
        self.vertices = [seed_vertex]
        self.weight = [{0: 0.0}]
        self.total_weight = 0.0

    def factory_add_vertex(self, vertex, weights_to_other):
        # Return a new graph as a copy of this with an extra vertex added. This
        # is a factory rather than a change in-place because CompleteGraph ought
        # to be immutable to implement __hash__
        g = copy.deepcopy(self)

        current_len = len(g.vertices)
        assert len(weights_to_other) == current_len
        g.vertices.append(vertex)
        node = current_len

        # Update distances from other nodes to the new one
        for i, w in enumerate(weights_to_other):
            g.weight[i][node] = w

        # Add distances to other nodes from this one
        weights_to_other.append(0.0)
        to_other = {}
        for i, w in enumerate(weights_to_other):
            to_other[i] = w
        g.weight.append(to_other)

        # Update the total weight
        g.total_weight += sum(weights_to_other)

        # Sort the vertices and weights by spot_id
        l = zip(g.vertices, g.weight)
        l = sorted(l, key=lambda v_w: v_w[0]["spot_id"])
        v, w = zip(*l)
        g.vertices = list(v)
        g.weight = list(w)

        return g

    def __hash__(self):
        h = tuple((e["spot_id"], e["miller_index"]) for e in self.vertices)
        return hash(h)

    def __eq__(self, other):
        for a, b in zip(self.vertices, other.vertices):
            if a["spot_id"] != b["spot_id"]:
                return False
            if a["miller_index"] != b["miller_index"]:
                return False
        return True

    def __ne__(self, other):
        return not self == other


low_res_spot_match_phil_str = """\
candidate_spots
{
    limit_resolution_by = *n_spots d_min
    .type = choice

    d_min = 15.0
    .type = float(value_min=0)

    n_spots = 10
    .type = int

    d_star_tolerance = 4.0
    .help = "Number of sigmas from the centroid position for which to "
            "calculate d* bands"
    .type = float
}

use_P1_indices_as_seeds = False
    .type = bool

search_depth = *triplets quads
    .type = choice

bootstrap_crystal = False
    .type = bool

max_pairs = 200
    .type = int

max_triplets = 600
    .type = int

max_quads = 600
    .type = int
"""


class LowResSpotMatch(Strategy):
    """
    A lattice search strategy matching low resolution spots to candidate indices.

    The match is based on resolution and reciprocal space distance between observed
    spots. A prior unit cell and space group are required and solutions are assessed
    by matching the low resolution spots against candidate reflection positions
    predicted from the known cell. This lattice search strategy is a special case
    designed to work for electron diffraction still images, in which one typically
    only collects reflections from the zero-order Laue zone. In principle, it is not
    limited to this type of data, but probably works best with narrow wedges,
    good initial geometry and a small beam-stop shadow so that a good number of
    low-order reflections are collected.
    """

    phil_help = (
        "A lattice search strategy that matches low resolution spots to candidate "
        "indices based on a known unit cell and space group. Designed primarily for "
        "electron diffraction still images."
    )

    phil_scope = libtbx.phil.parse(low_res_spot_match_phil_str)

    def __init__(
        self, target_symmetry_primitive, max_lattices, params=None, *args, **kwargs
    ):
        """Construct a LowResSpotMatch object.

        Args:
            target_symmetry_primitive (cctbx.crystal.symmetry): The target
                crystal symmetry and unit cell

            max_lattices (int): The maximum number of lattice models to find
        """
        super().__init__(params=params, *args, **kwargs)
        self._target_symmetry_primitive = target_symmetry_primitive
        self._max_lattices = max_lattices

        if target_symmetry_primitive is None:
            raise DialsIndexError(
                "Target unit cell and space group must be provided for low_res_spot_match"
            )

        # Set reciprocal space orthogonalisation matrix
        uc = self._target_symmetry_primitive.unit_cell()
        self.Bmat = matrix.sqr(uc.fractionalization_matrix()).transpose()

    def find_crystal_models(self, reflections, experiments):
        """Find a list of candidate crystal models.

        Args:
            reflections (dials.array_family.flex.reflection_table):
                The found spots centroids and associated data

            experiments (dxtbx.model.experiment_list.ExperimentList):
                The experimental geometry models
        """

        # Take a subset of the observations at the same resolution and calculate
        # some values that will be needed for the search
        self._calc_obs_data(reflections, experiments)

        # Construct a library of candidate low res indices with their d* values
        self._calc_candidate_hkls()

        # First search: match each observation with candidate indices within the
        # acceptable resolution band
        self._calc_seeds_and_stems()
        if self._params.use_P1_indices_as_seeds:
            seeds = self.stems
        else:
            seeds = self.seeds
        logger.info("Using %s seeds", len(seeds))

        # Second search: match seed spots with another spot from a different
        # reciprocal lattice row, such that the observed reciprocal space distances
        # are within tolerances
        pairs = []
        for seed in seeds:
            pairs.extend(self._pairs_with_seed(seed))
        logger.info("Found %s pairs", len(pairs))
        pairs = list(set(pairs))  # filter duplicates

        if self._params.max_pairs:
            pairs.sort(key=operator.attrgetter("total_weight"))
            idx = self._params.max_pairs
            pairs = pairs[0:idx]
        logger.info("Using %s highest-scoring pairs", len(pairs))

        # Further search iterations: extend to more spots within tolerated distances
        triplets = []
        for pair in pairs:
            triplets.extend(self._extend_by_candidates(pair))
        logger.info("Found %s triplets", len(triplets))
        triplets = list(set(triplets))  # filter duplicates
        if self._params.max_triplets:
            triplets.sort(key=operator.attrgetter("total_weight"))
            idx = self._params.max_triplets
            triplets = triplets[0:idx]
        logger.info("Using %s highest-scoring triplets", len(triplets))

        branches = triplets
        if self._params.search_depth == "quads":
            quads = []
            for triplet in triplets:
                quads.extend(self._extend_by_candidates(triplet))
            logger.info("%s quads", len(quads))
            quads = list(set(quads))  # filter duplicates
            if self._params.max_quads:
                quads.sort(key=operator.attrgetter("total_weight"))
                idx = self._params.max_quads
                quads = quads[0:idx]
            logger.info("Using %s highest-scoring quads", len(quads))
            branches = quads

        # Sort branches by total deviation of observed distances from expected
        branches.sort(key=operator.attrgetter("total_weight"))

        candidate_crystal_models = []
        for branch in branches:
            model = self._fit_crystal_model(branch)
            if model:
                candidate_crystal_models.append(model)
            if len(candidate_crystal_models) == self._max_lattices:
                break

        self.candidate_crystal_models = candidate_crystal_models
        return self.candidate_crystal_models

    def _calc_candidate_hkls(self):
        # First a list of indices that fill 1 ASU
        hkl_list = miller.build_set(
            self._target_symmetry_primitive,
            anomalous_flag=False,
            d_min=self._params.candidate_spots.d_min,
        )
        rt = flex.reflection_table()
        rt["miller_index"] = hkl_list.indices()
        rt["d_star"] = 1.0 / hkl_list.d_spacings().data()
        rt["rlp_datum"] = self.Bmat.elems * rt["miller_index"].as_vec3_double()
        self.candidate_hkls = rt

        # Now P1 indices with separate Friedel pairs
        hkl_list = miller.build_set(
            self._target_symmetry_primitive,
            anomalous_flag=True,
            d_min=self._params.candidate_spots.d_min,
        )
        hkl_list_p1 = hkl_list.expand_to_p1()
        rt = flex.reflection_table()
        rt["miller_index"] = hkl_list_p1.indices()
        rt["d_star"] = 1.0 / hkl_list_p1.d_spacings().data()
        rt["rlp_datum"] = self.Bmat.elems * rt["miller_index"].as_vec3_double()
        self.candidate_hkls_p1 = rt
        return

    def _calc_obs_data(self, reflections, experiments):
        """Calculates a set of low resolution observations to try to match to
        indices. Each observation will record its d* value as well as
        tolerated d* bands and a 'clock angle'"""

        spot_d_star = reflections["rlp"].norms()
        if self._params.candidate_spots.limit_resolution_by == "n_spots":
            n_spots = self._params.candidate_spots.n_spots
            n_spots = min(n_spots, len(reflections) - 1)
            d_star_max = flex.sorted(spot_d_star)[n_spots - 1]
            self._params.candidate_spots.d_min = 1.0 / d_star_max

        # First select low resolution spots only
        spot_d_star = reflections["rlp"].norms()
        d_star_max = 1.0 / self._params.candidate_spots.d_min
        sel = spot_d_star <= d_star_max
        self.spots = reflections.select(sel)
        self.spots["d_star"] = spot_d_star.select(sel)

        # XXX In what circumstance might there be more than one experiment?
        detector = experiments.detectors()[0]
        beam = experiments.beams()[0]

        # Lab coordinate of the beam centre, using the first spot's panel
        panel = detector[self.spots[0]["panel"]]
        bc = panel.get_ray_intersection(beam.get_s0())
        bc_lab = panel.get_lab_coord(bc)

        # Lab coordinate of each spot
        spot_lab = flex.vec3_double(len(self.spots))
        pnl_ids = set(self.spots["panel"])
        for pnl in pnl_ids:
            sel = self.spots["panel"] == pnl
            panel = detector[pnl]
            obs = self.spots["xyzobs.mm.value"].select(sel)
            x_mm, y_mm, _ = obs.parts()
            spot_lab.set_selected(
                sel, panel.get_lab_coord(flex.vec2_double(x_mm, y_mm))
            )

        # Radius vectors for each spot
        radius = spot_lab - bc_lab

        # Usually the radius vectors would all be in a single plane, but this might
        # not be the case if the spots are on different panels. To put them on the
        # same plane, project onto fast/slow of the panel used to get the beam
        # centre
        df = flex.vec3_double(len(self.spots), detector[0].get_fast_axis())
        ds = flex.vec3_double(len(self.spots), detector[0].get_slow_axis())
        clock_dirs = (radius.dot(df) * df + radius.dot(ds) * ds).each_normalize()

        # From this, find positive angles of each vector around a clock, using the
        # fast axis as 12 o'clock
        angs = clock_dirs.angle(detector[0].get_fast_axis())
        dots = clock_dirs.dot(detector[0].get_slow_axis())
        sel = dots < 0  # select directions in the second half of the clock face
        angs.set_selected(sel, (TWO_PI - angs.select(sel)))
        self.spots["clock_angle"] = angs

        # Project radius vectors onto fast/slow of the relevant panels
        df = flex.vec3_double(len(self.spots))
        ds = flex.vec3_double(len(self.spots))
        for pnl in pnl_ids:
            sel = self.spots["panel"] == pnl
            panel = detector[pnl]
            df.set_selected(sel, panel.get_fast_axis())
            ds.set_selected(sel, panel.get_slow_axis())
        panel_dirs = (radius.dot(df) * df + radius.dot(ds) * ds).each_normalize()

        # Calc error along each panel direction with simple error propagation
        # that assumes no covariance between x and y centroid errors.
        x = panel_dirs.dot(df)
        y = panel_dirs.dot(ds)
        x2, y2 = flex.pow2(x), flex.pow2(y)
        r2 = x2 + y2
        sig_x2, sig_y2, _ = self.spots["xyzobs.mm.variance"].parts()
        var_r = (x2 / r2) * sig_x2 + (y2 / r2) * sig_y2
        sig_r = flex.sqrt(var_r)

        # Pixel coordinates at limits of the band
        tol = self._params.candidate_spots.d_star_tolerance
        outer_spot_lab = spot_lab + panel_dirs * (tol * sig_r)
        inner_spot_lab = spot_lab - panel_dirs * (tol * sig_r)

        # Set d* at band limits
        inv_lambda = 1.0 / beam.get_wavelength()
        s1_outer = outer_spot_lab.each_normalize() * inv_lambda
        s1_inner = inner_spot_lab.each_normalize() * inv_lambda
        self.spots["d_star_outer"] = (s1_outer - beam.get_s0()).norms()
        self.spots["d_star_inner"] = (s1_inner - beam.get_s0()).norms()
        self.spots["d_star_band2"] = flex.pow2(
            self.spots["d_star_outer"] - self.spots["d_star_inner"]
        )

    def _calc_seeds_and_stems(self):
        # As the first stage of search, determine a list of seed spots for further
        # stages. Order these by distance of observed d* from the candidate
        # reflection's canonical d*

        # First the 'seeds' (in 1 ASU)
        self.seeds = []
        for i, spot in enumerate(self.spots.rows()):
            sel = (self.candidate_hkls["d_star"] <= spot["d_star_outer"]) & (
                self.candidate_hkls["d_star"] >= spot["d_star_inner"]
            )
            cands = self.candidate_hkls.select(sel)
            for c in cands.rows():
                r_dst = abs(c["d_star"] - spot["d_star"])
                self.seeds.append(
                    {
                        "spot_id": i,
                        "miller_index": c["miller_index"],
                        "rlp_datum": matrix.col(c["rlp_datum"]),
                        "residual_d_star": r_dst,
                        "clock_angle": spot["clock_angle"],
                    }
                )

        self.seeds.sort(key=operator.itemgetter("residual_d_star"))

        # Now the 'stems' to use in second search level, using all indices in P 1
        self.stems = []
        for i, spot in enumerate(self.spots.rows()):
            sel = (self.candidate_hkls_p1["d_star"] <= spot["d_star_outer"]) & (
                self.candidate_hkls_p1["d_star"] >= spot["d_star_inner"]
            )
            cands = self.candidate_hkls_p1.select(sel)
            for c in cands.rows():
                r_dst = abs(c["d_star"] - spot["d_star"])
                self.stems.append(
                    {
                        "spot_id": i,
                        "miller_index": c["miller_index"],
                        "rlp_datum": matrix.col(c["rlp_datum"]),
                        "residual_d_star": r_dst,
                        "clock_angle": spot["clock_angle"],
                    }
                )

        self.stems.sort(key=operator.itemgetter("residual_d_star"))

    def _pairs_with_seed(self, seed):
        seed_rlp = matrix.col(self.spots[seed["spot_id"]]["rlp"])

        result = []
        for cand in self.stems:
            # Don't check the seed spot itself
            if cand["spot_id"] == seed["spot_id"]:
                continue

            # Skip spots at a very similar clock angle, which probably belong to the
            # same line of indices from the origin
            angle_diff = cand["clock_angle"] - seed["clock_angle"]
            angle_diff = abs(((angle_diff + math.pi) % TWO_PI) - math.pi)
            if angle_diff < FIVE_DEG:
                continue

            # Calculate the plane normal for the plane containing the seed and stem.
            # Skip pairs of Miller indices that belong to the same line
            seed_vec = seed["rlp_datum"]
            cand_vec = cand["rlp_datum"]
            try:
                seed_vec.cross(cand_vec).normalize()
            except ZeroDivisionError:
                continue

            # Compare expected reciprocal space distance with observed distance
            cand_rlp = matrix.col(self.spots[cand["spot_id"]]["rlp"])
            obs_dist = (cand_rlp - seed_rlp).length()
            exp_dist = (seed_vec - cand_vec).length()
            r_dist = abs(obs_dist - exp_dist)

            # If the distance difference is larger than the sum in quadrature of the
            # tolerated d* bands then reject the candidate
            sq_band1 = self.spots[seed["spot_id"]]["d_star_band2"]
            sq_band2 = self.spots[cand["spot_id"]]["d_star_band2"]
            if r_dist > math.sqrt(sq_band1 + sq_band2):
                continue

            # Store the seed-stem match as a 2-node graph
            g = CompleteGraph(
                {
                    "spot_id": seed["spot_id"],
                    "miller_index": seed["miller_index"],
                    "rlp_datum": seed["rlp_datum"],
                }
            )
            g = g.factory_add_vertex(
                {
                    "spot_id": cand["spot_id"],
                    "miller_index": cand["miller_index"],
                    "rlp_datum": cand["rlp_datum"],
                },
                weights_to_other=[r_dist],
            )
            result.append(g)
        return result

    def _extend_by_candidates(self, graph):

        existing_ids = [e["spot_id"] for e in graph.vertices]
        obs_relps = [matrix.col(self.spots[e]["rlp"]) for e in existing_ids]
        exp_relps = [e["rlp_datum"] for e in graph.vertices]

        result = []

        for cand in self.stems:
            # Don't check spots already matched
            if cand["spot_id"] in existing_ids:
                continue

            # Compare expected reciprocal space distances with observed distances
            cand_rlp = matrix.col(self.spots[cand["spot_id"]]["rlp"])
            cand_vec = cand["rlp_datum"]

            obs_dists = [(cand_rlp - rlp).length() for rlp in obs_relps]
            exp_dists = [(vec - cand_vec).length() for vec in exp_relps]

            residual_dist = [abs(a - b) for (a, b) in zip(obs_dists, exp_dists)]

            # If any of the distance differences is larger than the sum in quadrature
            # of the tolerated d* bands then reject the candidate
            sq_candidate_band = self.spots[cand["spot_id"]]["d_star_band2"]
            bad_candidate = False
            for r_dist, spot_id in zip(residual_dist, existing_ids):
                sq_relp_band = self.spots[spot_id]["d_star_band2"]
                if r_dist > math.sqrt(sq_relp_band + sq_candidate_band):
                    bad_candidate = True
                    break
            if bad_candidate:
                continue

            # Calculate co-planarity of the relps, including the origin
            points = flex.vec3_double(exp_relps + [cand_vec, (0.0, 0.0, 0.0)])
            plane = least_squares_plane(points)
            plane_score = flex.sum_sq(
                points.dot(plane.normal) - plane.distance_to_origin
            )

            # Reject if the group of relps are too far from lying in a single plane.
            # This cut-off was determined by trial and error using simulated images.
            if plane_score > 6e-7:
                continue

            # Construct a graph including the accepted candidate node
            g = graph.factory_add_vertex(
                {
                    "spot_id": cand["spot_id"],
                    "miller_index": cand["miller_index"],
                    "rlp_datum": cand["rlp_datum"],
                },
                weights_to_other=residual_dist,
            )

            result.append(g)

        return result

    @staticmethod
    def _fit_U_from_superposed_points(reference, other):

        # Add the origin to both sets of points
        reference.append((0, 0, 0))
        other.append((0, 0, 0))

        # Find U matrix that takes ideal relps to the reference
        fit = superpose.least_squares_fit(reference, other)
        return fit.r

    def _fit_crystal_model(self, graph):

        vertices = graph.vertices

        # Reciprocal lattice points of the observations
        sel = flex.size_t([e["spot_id"] for e in vertices])
        reference = self.spots["rlp"].select(sel)

        # Ideal relps from the known cell
        other = flex.vec3_double([e["rlp_datum"] for e in vertices])

        U = self._fit_U_from_superposed_points(reference, other)
        UB = U * self.Bmat

        if self._params.bootstrap_crystal:

            # Attempt to index the low resolution spots
            from dials_algorithms_indexing_ext import AssignIndices

            phi = self.spots["xyzobs.mm.value"].parts()[2]
            UB_matrices = flex.mat3_double([UB])
            result = AssignIndices(self.spots["rlp"], phi, UB_matrices, tolerance=0.3)
            hkl = result.miller_indices()
            sel = hkl != (0, 0, 0)
            hkl_vec = hkl.as_vec3_double().select(sel)

            # Use the result to get a new UB matrix
            reference = self.spots["rlp"].select(sel)
            other = self.Bmat.elems * hkl_vec
            U = self._fit_U_from_superposed_points(reference, other)
            UB = U * self.Bmat

        # Calculate RMSD of the fit
        rms = reference.rms_difference(U.elems * other)

        # Construct a crystal model
        xl = Crystal(A=UB, space_group_symbol="P1")

        # Monkey-patch crystal to return rms of the fit (useful?)
        xl.rms = rms

        return xl



from __future__ import annotations

import os
import sys

import pytest

import dials.util


class _SomeError(RuntimeError):
    pass


def test_crash_with_plain_text():
    with pytest.raises(_SomeError) as exc:
        with dials.util.show_mail_on_error():
            raise _SomeError("This is a string")
    assert "report this error" in str(exc.value)
    assert "This is a string" in str(exc.value)


def test_crash_with_unicode():
    with pytest.raises(_SomeError) as exc:
        with dials.util.show_mail_on_error():
            raise _SomeError("This is a 👹♔  𝓕คｎｃ¥ ｕηเᶜόＤ𝔼  🏆🍔 string")
    assert "report this error" in str(exc.value)
    assert "This is a" in str(exc.value)


def test_crash_with_bytestring():
    with pytest.raises(_SomeError) as exc:
        with dials.util.show_mail_on_error():
            raise _SomeError(b"This is a byte string")
    assert "report this error" in repr(exc.value)
    assert "This is a" in repr(exc.value)


def test_coloured_exit(monkeypatch):
    with pytest.raises(SystemExit) as e:
        with dials.util.make_sys_exit_red():
            sys.exit("Ohno")
    assert e.value.code == "Ohno"

    class _AttySysErr:
        def isatty(self):
            return True

    # Force colouring on
    monkeypatch.setattr(sys, "stderr", _AttySysErr())

    with pytest.raises(SystemExit) as e:
        with dials.util.make_sys_exit_red():
            sys.exit("Ohno")
    assert str(e.value.code).startswith("\033[")

    # and off again...
    monkeypatch.setitem(os.environ, "NO_COLOR", "1")
    with pytest.raises(SystemExit) as e:
        with dials.util.make_sys_exit_red():
            sys.exit("Ohno")
    assert e.value.code == "Ohno"



from __future__ import annotations

import jinja2
from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.parsers.rst.directives import unchanged

BUTTON_TEMPLATE = jinja2.Template(
    '<a href="{{ link }}" class="button"><span>{{ text }}</span></a>'
)

# placeholder node for document graph
class button_node(nodes.General, nodes.Element):
    pass


class ButtonDirective(Directive):
    required_arguments = 0

    option_spec = {"text": unchanged, "link": unchanged}

    # this will execute when your directive is encountered
    # it will insert a button_node into the document that will
    # get visisted during the build phase
    def run(self):
        node = button_node()
        node["text"] = self.options["text"]
        node["link"] = self.options["link"]
        return [node]


# build phase visitor emits HTML to append to output
def html_visit_button_node(self, node):
    html = BUTTON_TEMPLATE.render(text=node["text"], link=node["link"])
    self.body.append(html)
    raise nodes.SkipNode


# if you want to be pedantic, define text, latex, manpage visitors too..


def setup(app):
    app.add_node(button_node, html=(html_visit_button_node, None))
    app.add_directive("button", ButtonDirective)
    return {"parallel_read_safe": True}



from __future__ import annotations

import math
import os
import random

import pytest

from dxtbx.model.experiment_list import ExperimentListFactory
from scitbx import matrix

import dials.util.nexus
from dials.command_line.export import phil_scope
from dials.util.nexus.nx_mx import (
    polarization_normal_to_stokes,
    polarization_stokes_to_normal,
)


def test_polarization_conversion():
    def conv(n, p):
        S0, S1, S2, S3 = polarization_normal_to_stokes(n, p)
        return polarization_stokes_to_normal(S0, S1, S2, S3)

    for i in range(1000):
        n1 = matrix.col((random.uniform(-1, 1), random.uniform(-1, 1), 0)).normalize()
        p1 = random.uniform(0, 1)
        n2, p2 = conv(n1, p1)
        angle = n2.angle(n1)
        assert (
            abs(angle) == pytest.approx(0, abs=1e-7)
            or angle == pytest.approx(math.pi)
            or angle == pytest.approx(math.pi * 2)
        )
        assert p1 == pytest.approx(p2, abs=1e-7)


def run_single(experiments1, filename):
    # Dump the file
    params = phil_scope.extract().nxs
    params.hklout = filename
    dials.util.nexus.dump(experiments1, None, params)

    # Load the file
    experiments2, reflections = dials.util.nexus.load(filename)
    assert experiments2 is not None
    assert reflections is None
    assert len(experiments2) == len(experiments1)

    index1 = []
    index2 = []

    for exp1, exp2 in zip(experiments1, experiments2):

        # Check the beam
        b1 = exp1.beam
        b2 = exp2.beam
        assert all(
            d1 == pytest.approx(d2)
            for d1, d2 in zip(
                b1.get_sample_to_source_direction(), b2.get_sample_to_source_direction()
            )
        )
        assert b1.get_wavelength() == pytest.approx(b2.get_wavelength())
        assert b1.get_polarization_fraction() == pytest.approx(
            b2.get_polarization_fraction()
        )
        n1 = matrix.col(b1.get_polarization_normal())
        n2 = matrix.col(b2.get_polarization_normal())
        angle = n2.angle(n1)
        assert (
            angle == pytest.approx(0, abs=1e-7)
            or angle == pytest.approx(math.pi)
            or angle == pytest.approx(2 * math.pi)
        )

        # Check the goniometer
        g1 = exp1.goniometer
        g2 = exp2.goniometer
        if g1 is None:
            assert g2 is None
        else:
            assert all(
                d1 == pytest.approx(d2)
                for d1, d2 in zip(g1.get_rotation_axis(), g2.get_rotation_axis())
            )
            assert all(
                d1 == pytest.approx(d2)
                for d1, d2 in zip(g1.get_fixed_rotation(), g2.get_fixed_rotation())
            )
            assert all(
                d1 == pytest.approx(d2)
                for d1, d2 in zip(g1.get_setting_rotation(), g2.get_setting_rotation())
            )

        # Check the scan
        s1 = exp1.scan
        s2 = exp2.scan
        if s1 is None:
            assert s2 is None
        else:
            assert len(s1) == len(s2)
            assert s1.get_image_range() == s2.get_image_range()
            assert s1.get_oscillation() == pytest.approx(s2.get_oscillation())
            assert s1.get_exposure_times() == pytest.approx(s2.get_exposure_times())
            assert s1.get_epochs() == pytest.approx(s2.get_epochs())

        # Check the detector
        d1 = exp1.detector
        d2 = exp2.detector
        assert len(d1) == len(d2)
        for p1, p2 in zip(d1, d2):
            assert p1.get_type() == p2.get_type()
            assert p1.get_material() == p2.get_material()
            assert p1.get_thickness() == p2.get_thickness()
            assert p1.get_image_size() == p2.get_image_size()
            assert p1.get_pixel_size() == p2.get_pixel_size()
            assert p1.get_trusted_range() == p2.get_trusted_range()
            assert p1.get_fast_axis() == pytest.approx(p2.get_fast_axis())
            assert p1.get_slow_axis() == pytest.approx(p2.get_slow_axis())
            assert p1.get_origin() == pytest.approx(p2.get_origin())

        # Check the crystal
        c1 = exp1.crystal
        c2 = exp2.crystal
        assert c1.get_space_group() == c2.get_space_group()
        assert c1.get_unit_cell().parameters() == pytest.approx(
            c2.get_unit_cell().parameters()
        )
        assert c1.get_A() == pytest.approx(c2.get_A())
        assert c1.num_scan_points == c2.num_scan_points
        for i in range(c1.num_scan_points):
            assert c1.get_A_at_scan_point(i) == pytest.approx(c2.get_A_at_scan_point(i))
            uc1 = c1.get_unit_cell_at_scan_point(i)
            uc2 = c2.get_unit_cell_at_scan_point(i)
            assert uc1.parameters() == pytest.approx(uc2.parameters())

        index1.append(
            (
                id(exp1.beam),
                id(exp1.detector),
                id(exp1.goniometer),
                id(exp1.scan),
                id(exp1.crystal),
            )
        )

        index2.append(
            (
                id(exp2.beam),
                id(exp2.detector),
                id(exp2.goniometer),
                id(exp2.scan),
                id(exp2.crystal),
            )
        )

    # Get a list of all beam etc
    beam1, detector1, goniometer1, scan1, crystal1 = zip(*index1)
    beam2, detector2, goniometer2, scan2, crystal2 = zip(*index2)

    # If any models are shared then check they are shared in both
    num = len(beam1)
    for i in range(0, num - 1):
        for j in range(1, num):
            assert (beam1[i] == beam1[j]) is (beam2[i] == beam2[j])
            assert (detector1[i] == detector1[j]) is (detector2[i] == detector2[j])
            assert (goniometer1[i] == goniometer1[j]) is (
                goniometer2[i] == goniometer2[j]
            )
            assert (scan1[i] == scan1[j]) is (scan2[i] == scan2[j])
            assert (crystal1[i] == crystal1[j]) is (crystal2[i] == crystal2[j])


@pytest.mark.parametrize(
    "filename",
    [
        "single",
        "multiple_unrelated",
        "multi_crystal",
        "two_colour",
        "multiple_sweeps",
        "stills",
    ],
)
def test_nexus_dump_and_reload(dials_regression, tmpdir, filename):
    path = os.path.join(dials_regression, "nexus_test_data", "shared_models")
    filename_in = os.path.join(path, f"{filename}.json")
    filename_out = tmpdir.join(f"{filename}.nxs").strpath
    experiments = ExperimentListFactory.from_json_file(filename_in)
    run_single(experiments, filename_out)



from __future__ import annotations

import os
import pickle

import wx
import wx.lib.colourselect

import rstbx.viewer.display
import wxtbx.plots
from libtbx.utils import to_unicode
from wxtbx import bitmaps, icons

from dials.util import Sorry

# Instance to bind external update event to an event handler
EVT_EXTERNAL_UPDATE = wx.PyEventBinder(wx.NewEventType(), 0)


class ExternalUpdateEvent(wx.PyCommandEvent):
    """XXX This class, along with the EVT_EXTERNAL_UPDATE instance
    should perhaps move into its own file?
    """

    def __init__(self, eventType=EVT_EXTERNAL_UPDATE.evtType[0], id=0):
        wx.PyCommandEvent.__init__(self, eventType, id)
        self.img = None
        self.title = None


class XrayFrame(wx.Frame):
    # Maximum number of entries in the chooser.
    CHOOSER_SIZE = 1024

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.settings = rstbx.viewer.settings()
        self.viewer = rstbx.viewer.display.XrayView(self, -1, size=(1024, 640))
        self.viewer.SetMinSize((640, 640))
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)
        self.sizer.Add(self.viewer, 1, wx.EXPAND)
        self.statusbar = self.CreateStatusBar()
        self.settings_frame = None
        self.zoom_frame = None
        self.zoom_3d = None
        self.plot_frame = None
        self._img = None
        self._distl = None
        self.toolbar = self.CreateToolBar(style=wx.TB_TEXT)
        self.setup_toolbar()
        self.toolbar.Realize()
        self.mb = wx.MenuBar()
        self.setup_menus()
        self.SetMenuBar(self.mb)
        self.Fit()
        self.SetMinSize(self.GetSize())
        self.OnShowSettings(None)
        self.Bind(EVT_EXTERNAL_UPDATE, self.OnExternalUpdate)

    def OnExternalUpdate(self, event):
        """The OnExternalUpdate() function updates the image and the title
        from @p event.
        """

        # See self.load_image().
        self._img = event.img
        self.viewer.set_image(self._img)
        if self.settings_frame is not None:
            self.settings_frame.set_image(self._img)
        if self.zoom_frame is not None:
            self.zoom_frame.set_image(self._img)
        self.SetTitle(event.title)
        self.update_statusbar()
        self.Layout()

    def setup_toolbar(self):
        btn = self.toolbar.AddTool(
            toolId=-1,
            label="Load file",
            bitmap=icons.hkl_file.GetBitmap(),
            shortHelp="Load file",
            kind=wx.ITEM_NORMAL,
        )
        self.Bind(wx.EVT_MENU, self.OnLoadFile, btn)
        btn = self.toolbar.AddTool(
            toolId=-1,
            label="Settings",
            bitmap=icons.advancedsettings.GetBitmap(),
            shortHelp="Settings",
            kind=wx.ITEM_NORMAL,
        )
        self.Bind(wx.EVT_MENU, self.OnShowSettings, btn)
        btn = self.toolbar.AddTool(
            toolId=-1,
            label="Zoom",
            bitmap=icons.search.GetBitmap(),
            shortHelp="Zoom",
            kind=wx.ITEM_NORMAL,
        )
        self.Bind(wx.EVT_MENU, self.OnZoom, btn)
        txt = wx.StaticText(self.toolbar, -1, "Image:")
        self.toolbar.AddControl(txt)
        self.image_chooser = wx.Choice(self.toolbar, -1, size=(300, -1))
        self.toolbar.AddControl(self.image_chooser)
        self.Bind(wx.EVT_CHOICE, self.OnChooseImage, self.image_chooser)
        btn = self.toolbar.AddTool(
            toolId=wx.ID_BACKWARD,
            label="Previous",
            bitmap=bitmaps.fetch_icon_bitmap("actions", "1leftarrow"),
            shortHelp="Previous",
            kind=wx.ITEM_NORMAL,
        )
        self.Bind(wx.EVT_MENU, self.OnPrevious, btn)
        btn = self.toolbar.AddTool(
            toolId=wx.ID_FORWARD,
            label="Next",
            bitmap=bitmaps.fetch_icon_bitmap("actions", "1rightarrow"),
            shortHelp="Next",
            kind=wx.ITEM_NORMAL,
        )
        self.Bind(wx.EVT_MENU, self.OnNext, btn)

    def setup_menus(self):
        file_menu = wx.Menu()
        self.mb.Append(file_menu, "File")
        item = file_menu.Append(-1, "Open integration results...")
        self.Bind(wx.EVT_MENU, self.OnLoadIntegration, item)
        item = file_menu.Append(-1, "Open image...")
        self.Bind(wx.EVT_MENU, self.OnLoadFile, item)
        actions_menu = wx.Menu()
        self.mb.Append(actions_menu, "Actions")
        item = actions_menu.Append(-1, "Change beam center...")
        self.Bind(wx.EVT_MENU, self.OnChangeBeamCenter, item)
        item = actions_menu.Append(-1, "Reset beam center to header value")
        self.Bind(wx.EVT_MENU, lambda evt: self.viewer.ResetBeamCenter(), item)
        item = actions_menu.Append(-1, "Save screenshot...")
        self.Bind(wx.EVT_MENU, self.OnScreenShot, item)

    def get_key(self, file_name_or_data):
        """The get_key() function returns the key of @p file_name_or_data.
        In the case of dictionaries, it is the timestamp of the image.
        For file names, the key is an ASCII-encoded absolute path string.
        Otherwise, get_key() returns @c None.
        """

        try:
            return file_name_or_data["TIMESTAMP"]
        except TypeError:
            pass

        try:
            return file_name_or_data.get_image_file()
        except AttributeError:
            pass

        try:
            return os.path.abspath(file_name_or_data)
        except TypeError:
            pass

        try:
            return os.path.abspath(file_name_or_data.encode("ascii"))
        except (AttributeError, TypeError):
            pass

        return None

    def load_image(self, file_name_or_data):
        """The load_image() function displays the image from @p
        file_name_or_data.  The chooser is updated appropriately.
        """

        key = self.get_key(file_name_or_data)
        if type(file_name_or_data) is dict:
            self._img = rstbx.viewer.image(file_name_or_data)
        else:
            try:
                self._img = rstbx.viewer.image(key)
            except OSError:
                raise Sorry(
                    (
                        "The file '%s' could not be recognized as a supported "
                        + "image format; please make sure it is actually a detector image."
                    )
                    % key
                )

        # Update the selection in the chooser.
        i = self.add_file_name_or_data(file_name_or_data)
        self.image_chooser.SetSelection(i)

        self.viewer.set_image(self._img)
        if self.settings_frame is not None:
            self.settings_frame.set_image(self._img)
        self.SetTitle(to_unicode(key))
        self.update_statusbar()
        self.Layout()

    def load_distl_output(self, file_name):
        with open(file_name, "rb") as fh:
            distl = pickle.load(fh)
        self._distl = distl
        img_files = []
        for img_id in sorted(distl.images.keys()):
            img = distl.images[img_id]
            img_files.append(img["relpath"])
        if len(img_files) == 0:
            raise Sorry("No images in this result!")
        self.image_chooser.SetItems([os.path.basename(f) for f in img_files])
        self.image_chooser.SetSelection(0)
        self.load_image(img_files[0])
        self.annotate_image(img_files[0])

    def add_file_name_or_data(self, file_name_or_data):
        """The add_file_name_or_data() function appends @p
        file_name_or_data to the image chooser, unless it is already
        present.  For file-backed images, the base name is displayed in
        the chooser.  If necessary, the number of entries in the chooser
        is pruned.  The function returns the index of the recently added
        entry.  XXX This is probably the place for heuristics to determine
        if the viewer was given a pattern, or a plain list of files.  XXX
        Rename this function, because it only deals with the chooser?
        """

        key = self.get_key(file_name_or_data)
        for i in range(self.image_chooser.GetCount()):
            if key == self.image_chooser.GetClientData(i):
                return i
        if self.image_chooser.GetCount() >= self.CHOOSER_SIZE:
            self.image_chooser.Delete(0)
        i = self.image_chooser.GetCount()
        if type(file_name_or_data) is dict:
            self.image_chooser.Insert(key, i, None)
        else:
            self.image_chooser.Insert(os.path.basename(key), i, key)
        return i

    def annotate_image(self, file_name):
        assert self._distl is not None
        for img_id in sorted(self._distl.images.keys()):
            img = self._distl.images[img_id]
            if img["relpath"] == file_name:
                spots = img["spotoutput"]["inlier_spots"]
                self._img.set_spots(spots)
                break

    def load_integration(self, dir_name):
        from rstbx.viewer import processing

        assert os.path.isdir(dir_name)
        self.proc_frame = processing.ProcessingFrame(None, -1, "LABELIT")
        self.proc_frame.set_viewer_frame(self)
        self.proc_frame.LoadResults(dir_name)
        self.proc_frame.Show()

    def display_integration_result(self, result):
        assert isinstance(result, dict)
        self._img.set_integration_results(result)
        self.viewer.Refresh()

    def update_statusbar(self, info=None):
        if info is None:
            self.statusbar.SetStatusText(
                "Click and drag to pan, mouse wheel or double click to zoom"
            )
        else:
            self.statusbar.SetStatusText(info.format())

    def update_settings(self, layout=True):
        self.viewer.update_settings(layout)

    def set_brightness(self, brightness):
        if (brightness > 0) and (brightness <= 500):
            self.settings.brightness = brightness
            if self.settings_frame is not None:
                self.settings_frame.update_controls()
            self.viewer.update_settings(layout=False)

    def OnLoadFile(self, event):
        wildcard_str = ""
        if wx.PlatformInfo[4] != "wxOSX-cocoa":
            from iotbx import file_reader

            wildcard_str = file_reader.get_wildcard_string("img")
        file_name = wx.FileSelector(
            "Image file",
            wildcard=wildcard_str,
            default_path="",
            flags=wx.FD_OPEN,
        )
        if file_name != "":
            self.load_image(file_name)

    def OnLoadLabelitResult(self, event):
        file_name = wx.FileSelector("Labelit result", default_path="", flags=wx.FD_OPEN)
        if file_name != "":
            self.load_image(file_name)

    def OnLoadIntegration(self, event):
        dir_name = wx.DirSelector("Integration result", defaultPath="")
        if dir_name != "":
            self.load_integration(dir_name)

    def OnShowSettings(self, event):
        raise NotImplementedError("Removed due to non-used code path")

    def OnShowZoom(self, event):
        if self.zoom_frame is None:
            self.zoom_frame = ZoomFrame(
                self,
                -1,
                "Zoom",
                style=wx.CAPTION | wx.CLOSE_BOX | wx.RESIZE_BORDER | wx.SYSTEM_MENU,
            )
            self.zoom_frame.set_image(self._img)
            self.zoom_frame.Show()
        self.zoom_frame.Raise()

    def OnShow3D(self, event):
        if self.zoom_3d is None:
            from rstbx.viewer import pixels3d

            self.zoom_3d = pixels3d.pixel_viewer_3d_frame(
                self,
                -1,
                "3D view",
                style=wx.CAPTION | wx.CLOSE_BOX | wx.RESIZE_BORDER | wx.SYSTEM_MENU,
            )
            self.zoom_3d.set_image(self._img)
            self.zoom_3d.Show()
        self.zoom_3d.Raise()

    def OnShowPlot(self, event):
        if self.plot_frame is None:
            self.plot_frame = PlotFrame(
                self,
                -1,
                "Intensity profile",
                style=wx.CAPTION | wx.CLOSE_BOX | wx.SYSTEM_MENU,
            )
            self.plot_frame.Show()

    def OnZoom(self, event):
        if self.settings.zoom_level == 6:
            self.settings.zoom_level = 0
        else:
            self.settings.zoom_level += 1
        if hasattr(self.viewer, "update_settings"):
            self.viewer.update_settings(layout=True)
        if self.settings_frame is not None:
            self.settings_frame.update_controls()

    def OnChooseImage(self, event):
        self.load_image(
            self.image_chooser.GetClientData(self.image_chooser.GetSelection())
        )

    def OnPrevious(self, event):
        n = self.image_chooser.GetSelection()
        if n != wx.NOT_FOUND and n - 1 >= 0:
            self.load_image(self.image_chooser.GetClientData(n - 1))

    def OnNext(self, event):
        n = self.image_chooser.GetSelection()
        if n != wx.NOT_FOUND and n + 1 < self.image_chooser.GetCount():
            self.load_image(self.image_chooser.GetClientData(n + 1))

    def OnScreenShot(self, event):
        file_name = wx.FileSelector(
            default_filename="xray.png",
            default_path="",
            wildcard="PNG image (*.png)|*.png",
            flags=wx.SAVE,
        )
        if file_name != "":
            self.viewer.save_image(file_name)

    def OnChangeBeamCenter(self, event):
        wx.MessageBox("Click on any point in the image to set the new beam center.")
        self.statusbar.SetStatusText("Changing beam center")
        self.viewer.ChangeBeamCenter()


mag_levels = [8, 16, 24, 32, 48, 64]


class ZoomFrame(wx.MiniFrame):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.settings = self.GetParent().settings
        self.control_panel = wx.Panel(self)
        self.panel = rstbx.viewer.display.ZoomView(self, -1)
        szr = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(szr)
        szr.Add(self.control_panel)
        szr.Add(self.panel, 1, wx.EXPAND)
        self.numbers_box = wx.CheckBox(self.control_panel, -1, "Show intensity values")
        txt1 = wx.StaticText(self.control_panel, -1, "Text color:")
        self.text_color = wx.lib.colourselect.ColourSelect(
            self.control_panel, colour=(255, 255, 0)
        )
        pszr = wx.BoxSizer(wx.VERTICAL)
        self.control_panel.SetSizer(pszr)
        box1 = wx.BoxSizer(wx.HORIZONTAL)
        pszr.Add(box1)
        box1.Add(self.numbers_box, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box1.Add(txt1, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box1.Add(self.text_color, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box2 = wx.BoxSizer(wx.HORIZONTAL)
        pszr.Add(box2)
        txt2 = wx.StaticText(self.control_panel, -1, "Magnification:")
        self.mag_ctrl = wx.Choice(
            self.control_panel, -1, choices=["%dx" % x for x in mag_levels]
        )
        self.mag_ctrl.SetSelection(1)
        box2.Add(txt2, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box2.Add(self.mag_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(wx.EVT_CLOSE, lambda evt: self.Destroy(), self)
        self.Bind(wx.EVT_WINDOW_DESTROY, self.OnDestroy)
        self.Bind(wx.EVT_CHECKBOX, self.OnChangeSettings, self.numbers_box)
        self.Bind(
            wx.lib.colourselect.EVT_COLOURSELECT, self.OnChangeSettings, self.text_color
        )
        self.Bind(wx.EVT_CHOICE, self.OnChangeSettings, self.mag_ctrl)
        szr.Fit(self.panel)
        self.Fit()

    def __getattr__(self, name):
        return getattr(self.panel, name)

    def OnDestroy(self, event):
        self.GetParent().zoom_frame = None

    def OnChangeSettings(self, event):
        self.panel.flag_show_intensities = self.numbers_box.GetValue()
        self.panel.text_color = self.text_color.GetValue()
        zoom = mag_levels[self.mag_ctrl.GetSelection()]
        self.panel.zoom_level = zoom
        self.Refresh()


class PlotFrame(wx.MiniFrame):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.plot = LinePlot(self, figure_size=(8, 3))
        szr = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(szr)
        szr.Add(self.plot, 1, wx.EXPAND)
        self.Fit()
        self.Bind(wx.EVT_CLOSE, lambda evt: self.Destroy(), self)
        self.Bind(wx.EVT_WINDOW_DESTROY, self.OnDestroy)

    def __getattr__(self, name):
        return getattr(self.plot, name)

    def OnDestroy(self, event):
        self.GetParent().plot_frame = None


class LinePlot(wxtbx.plots.plot_container):
    def show_plot(self, line):
        self.figure.clear()
        ax = self.figure.add_subplot(111)
        x_data = list(range(len(line.values)))
        ax.plot(x_data, line.values, "b-", linewidth=1)
        ax.set_ylabel("Intensity")
        if line.lattice_length is not None:
            ax.set_title(
                "Line distance = %.2fmm; avg. lattice length = %.2f Angstrom"
                % (line.distance, line.lattice_length)
            )
        else:
            ax.set_title(f"Line distance: {line.distance:.2f}mm")
        self.canvas.draw()



from __future__ import annotations

import os
import sys

from PIL import Image

import iotbx.phil
from dxtbx.model.detector_helpers import get_detector_projection_2d_axes

from dials.algorithms.image.threshold import DispersionThresholdDebug
from dials.array_family import flex
from dials.util import Sorry, show_mail_handle_errors
from dials.util.image_viewer.slip_viewer.flex_image import (
    get_flex_image,
    get_flex_image_multipanel,
)
from dials.util.options import ArgumentParser, flatten_experiments

help_message = """

Export raw diffraction image files as bitmap images, optionally exporting
images from intermediate spot-finding steps (local mean and variance maps,
or sigma_b, sigma_s or threshold-filtered images). Appearance of the images
can be altered via the brightness and colour_scheme parameters, and optionally
binning of pixels can be used to reduce image sizes.

Examples::

  dials.export_bitmaps image.cbf

  dials.export_bitmaps models.expt

  dials.export_bitmaps image.cbf display=variance colour_scheme=inverse_greyscale
"""

phil_scope = iotbx.phil.parse(
    """
binning = 1
  .type = int(value_min=1)
brightness = 100
  .type = float(value_min=0.0)
colour_scheme = *greyscale rainbow heatmap inverse_greyscale
  .type = choice
projection = lab *image
  .type = choice
padding = 4
  .type = int(value_min=0)
imageset_index = None
  .type = int
  .multiple = True
  .help = "The index/indices from an imageset to export. The first image of "
          "the set is 1."
  .expert_level=2
display = *image mean variance dispersion sigma_b \
          sigma_s threshold global_threshold
  .type = choice
nsigma_b = 6
  .type = float(value_min=0)
nsigma_s = 3
  .type = float(value_min=0)
global_threshold = 0
  .type = float(value_min=0)
kernel_size = 3,3
  .type = ints(size=2, value_min=1)
min_local = 2
  .type = int
gain = 1
  .type = float(value_min=0)
saturation = 0
  .type = int
show_mask = False
  .type = bool
png {
  compress_level = 1
    .type = int(value_min=0, value_max=9)
    .help = "ZLIB compression level, a number between 0 and 9: 1 gives best "
            "speed, 9 gives best compression, 0 gives no compression at all."
}
jpeg {
  quality = 75
    .type = int(value_min=1, value_max=95)
    .help = "The image quality, on a scale from 1 (worst) to 95 (best)"
}

include scope dials.util.options.format_phil_scope

output {
  prefix = "image"
    .type = str
  directory = None
    .type = path
  file = None
    .type = str
    .help = "Full name of the output file. Overrides 'prefix' and the default "
            "file extension. Only makes sense if a single file is written."
  format = jpeg *png tiff
    .type = choice
}""",
    process_includes=True,
)

colour_schemes = {"greyscale": 0, "rainbow": 1, "heatmap": 2, "inverse_greyscale": 3}


@show_mail_handle_errors()
def run(args=None):
    usage = "dials.export_bitmaps [options] models.expt | image.cbf"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        read_experiments_from_images=True,
        check_format=True,
        epilog=help_message,
    )

    params, options = parser.parse_args(args, show_diff_phil=True)

    experiments = flatten_experiments(params.input.experiments)
    if len(experiments) == 0:
        parser.print_help()
        exit(0)

    imagesets = experiments.imagesets()

    for imageset in imagesets:
        imageset_as_bitmaps(imageset, params)


def imageset_as_bitmaps(imageset, params):
    brightness = params.brightness / 100
    vendortype = "made up"
    # check that binning is a power of 2
    binning = params.binning
    if not (binning > 0 and ((binning & (binning - 1)) == 0)):
        raise Sorry("binning must be a power of 2")
    output_dir = params.output.directory
    if output_dir is None:
        output_dir = "."
    elif not os.path.exists(output_dir):
        os.makedirs(output_dir)
    output_files = []

    detector = imageset.get_detector()

    # Furnish detector with 2D projection axes
    detector.projected_2d = get_detector_projection_2d_axes(detector)
    detector.projection = params.projection

    panel = detector[0]
    scan = imageset.get_scan()
    # XXX is this inclusive or exclusive?
    saturation = panel.get_trusted_range()[1]
    if params.saturation:
        saturation = params.saturation
    if scan is not None and not scan.is_still() and not params.imageset_index:
        start, end = scan.get_image_range()
    else:
        start, end = 1, len(imageset)
    # If the user specified an image range index, only export those
    image_range = [
        i
        for i in range(start, end + 1)
        if not params.imageset_index or i in params.imageset_index
    ]
    if params.output.file and len(image_range) != 1:
        sys.exit("output.file can only be specified if a single image is exported")
    for i_image in image_range:
        image = imageset.get_raw_data(i_image - start)

        mask = imageset.get_mask(i_image - start)
        if mask is None:
            mask = [p.get_trusted_range_mask(im) for im, p in zip(image, detector)]

        if params.show_mask:
            for rd, m in zip(image, mask):
                rd.set_selected(~m, -2)

        image = image_filter(
            image,
            mask,
            display=params.display,
            gain_value=params.gain,
            nsigma_b=params.nsigma_b,
            nsigma_s=params.nsigma_s,
            global_threshold=params.global_threshold,
            min_local=params.min_local,
            kernel_size=params.kernel_size,
        )

        show_untrusted = params.show_mask
        if len(detector) > 1:
            # FIXME This doesn't work properly, as flex_image.size2() is incorrect
            # also binning doesn't work
            flex_image = get_flex_image_multipanel(
                brightness=brightness,
                detector=detector,
                image_data=image,
                binning=binning,
                beam=imageset.get_beam(),
                show_untrusted=show_untrusted,
            )
        else:
            flex_image = get_flex_image(
                brightness=brightness,
                data=image[0],
                binning=binning,
                saturation=saturation,
                vendortype=vendortype,
                show_untrusted=show_untrusted,
            )

        flex_image.setWindow(0, 0, 1)
        flex_image.adjust(color_scheme=colour_schemes.get(params.colour_scheme))

        # now export as a bitmap
        flex_image.prep_string()

        # XXX is size//binning safe here?
        pil_img = Image.frombytes(
            "RGB", (flex_image.ex_size2(), flex_image.ex_size1()), flex_image.as_bytes()
        )
        if params.output.file:
            path = os.path.join(output_dir, params.output.file)
        else:
            path = os.path.join(
                output_dir,
                "{prefix}{image:0{padding}}.{format}".format(
                    image=i_image,
                    prefix=params.output.prefix,
                    padding=params.padding,
                    format=params.output.format,
                ),
            )

        print(f"Exporting {path}")
        output_files.append(path)
        with open(path, "wb") as tmp_stream:
            pil_img.save(
                tmp_stream,
                format=params.output.format,
                compress_level=params.png.compress_level,
                quality=params.jpeg.quality,
            )

    return output_files


def image_filter(
    raw_data,
    mask,
    display,
    gain_value,
    nsigma_b,
    nsigma_s,
    global_threshold,
    min_local,
    kernel_size,
):

    if display == "image":
        return raw_data

    assert gain_value > 0
    gain_map = [flex.double(rd.accessor(), gain_value) for rd in raw_data]

    kabsch_debug_list = [
        DispersionThresholdDebug(
            data.as_double(),
            mask[i_panel],
            gain_map[i_panel],
            kernel_size,
            nsigma_b,
            nsigma_s,
            global_threshold,
            min_local,
        )
        for i_panel, data in enumerate(raw_data)
    ]

    if display == "mean":
        display_data = [kabsch.mean() for kabsch in kabsch_debug_list]
    elif display == "variance":
        display_data = [kabsch.variance() for kabsch in kabsch_debug_list]
    elif display == "dispersion":
        display_data = [kabsch.index_of_dispersion() for kabsch in kabsch_debug_list]
    elif display == "sigma_b":
        cv = [kabsch.index_of_dispersion() for kabsch in kabsch_debug_list]
        display_data = (kabsch.cv_mask() for kabsch in kabsch_debug_list)
        display_data = [_mask.as_1d().as_double() for _mask in display_data]
        for i, _mask in enumerate(display_data):
            _mask.reshape(cv[i].accessor())
    elif display == "sigma_s":
        cv = [kabsch.index_of_dispersion() for kabsch in kabsch_debug_list]
        display_data = (kabsch.value_mask() for kabsch in kabsch_debug_list)
        display_data = [_mask.as_1d().as_double() for _mask in display_data]
        for i, _mask in enumerate(display_data):
            _mask.reshape(cv[i].accessor())
    elif display == "global_threshold":
        cv = [kabsch.index_of_dispersion() for kabsch in kabsch_debug_list]
        display_data = (kabsch.global_mask() for kabsch in kabsch_debug_list)
        display_data = [_mask.as_1d().as_double() for _mask in display_data]
        for i, _mask in enumerate(display_data):
            _mask.reshape(cv[i].accessor())
    elif display == "threshold":
        cv = [kabsch.index_of_dispersion() for kabsch in kabsch_debug_list]
        display_data = (kabsch.final_mask() for kabsch in kabsch_debug_list)
        display_data = [_mask.as_1d().as_double() for _mask in display_data]
        for i, _mask in enumerate(display_data):
            _mask.reshape(cv[i].accessor())

    return display_data


if __name__ == "__main__":
    run()



from __future__ import annotations


class Test2d:
    def setup_class(self):
        from dials.algorithms.image.connected_components import LabelImageStack2d

        self.size = (500, 500)
        self.label_images = LabelImageStack2d(self.size)

        from scitbx.array_family import flex

        self.data_list = []
        self.mask_list = []
        for i in range(10):
            data = flex.random_int_gaussian_distribution(
                self.size[0] * self.size[1], 100, 10
            )
            data.reshape(flex.grid(self.size))
            mask = flex.random_bool(self.size[0] * self.size[1], 0.1)
            mask.reshape(flex.grid(self.size))
            self.data_list.append(data)
            self.mask_list.append(mask)

        for i in range(10):
            self.label_images.add_image(self.data_list[i], self.mask_list[i])

        self.labels = self.label_images.labels()
        self.coords = self.label_images.coords()
        self.values = list(self.label_images.values())

        assert len(self.labels) > 0
        assert len(self.labels) == len(self.coords)
        assert len(self.labels) == len(self.values)

    def test_coords_are_valid(self):
        # Ensure that the values are all ok and in the right order
        vi = 0
        for k in range(10):
            ind = 0
            for j in range(self.size[0]):
                for i in range(self.size[1]):
                    m = self.mask_list[k][ind]
                    if m:
                        c1 = (k, j, i)
                        c2 = self.coords[vi]
                        vi += 1
                        assert c1 == c2
                    ind += 1

    def test_values_are_valid(self):
        # Ensure that the values are all ok and in the right order
        vi = 0
        for k in range(10):
            for d, m in zip(self.data_list[k], self.mask_list[k]):
                if m:
                    v1 = d
                    v2 = self.values[vi]
                    vi += 1
                    assert v1 == v2

    def test_labels_are_valid(self):
        from scitbx.array_family import flex

        # Create a map of labels
        label_map = flex.int(flex.grid(10, self.size[0], self.size[1]))
        for c, l in zip(self.coords, self.labels):
            assert c[0] >= 0 and c[0] < 10
            assert c[1] >= 0 and c[1] < self.size[0]
            assert c[2] >= 0 and c[2] < self.size[1]
            label_map[c] = l

        # Ensure all labels are correct
        vi = 0
        for k in range(10):
            for j in range(self.size[0]):
                for i in range(self.size[1]):
                    if self.mask_list[k][j, i]:

                        l1 = self.labels[vi]
                        if k > 0 and self.mask_list[k - 1][j, i]:
                            l2 = label_map[k - 1, j, i]
                            assert l2 != l1
                        if j > 0 and self.mask_list[k][j - 1, i]:
                            l2 = label_map[k, j - 1, i]
                            assert l2 == l1
                        if i > 0 and self.mask_list[k][j, i - 1]:
                            l2 = label_map[k, j, i - 1]
                            assert l2 == l1
                        vi += 1


class Test3d:
    def setup_class(self):
        from dials.algorithms.image.connected_components import LabelImageStack3d

        self.size = (500, 500)
        self.label_images = LabelImageStack3d(self.size)

        from scitbx.array_family import flex

        self.data_list = []
        self.mask_list = []
        for i in range(10):
            data = flex.random_int_gaussian_distribution(
                self.size[0] * self.size[1], 100, 10
            )
            data.reshape(flex.grid(self.size))
            mask = flex.random_bool(self.size[0] * self.size[1], 0.1)
            mask.reshape(flex.grid(self.size))
            self.data_list.append(data)
            self.mask_list.append(mask)

        for i in range(10):
            self.label_images.add_image(self.data_list[i], self.mask_list[i])

        self.labels = self.label_images.labels()
        self.coords = self.label_images.coords()
        self.values = list(self.label_images.values())

        assert len(self.labels) > 0
        assert len(self.labels) == len(self.coords)
        assert len(self.labels) == len(self.values)

    def test_coords_are_valid(self):
        # Ensure that the values are all ok and in the right order
        vi = 0
        for k in range(10):
            ind = 0
            for j in range(self.size[0]):
                for i in range(self.size[1]):
                    m = self.mask_list[k][ind]
                    if m:
                        c1 = (k, j, i)
                        c2 = self.coords[vi]
                        vi += 1
                        assert c1 == c2
                    ind += 1

    def test_values_are_valid(self):
        # Ensure that the values are all ok and in the right order
        vi = 0
        for k in range(10):
            for d, m in zip(self.data_list[k], self.mask_list[k]):
                if m:
                    v1 = d
                    v2 = self.values[vi]
                    vi += 1
                    assert v1 == v2

    def test_labels_are_valid(self):
        from scitbx.array_family import flex

        # Create a map of labels
        label_map = flex.int(flex.grid(10, self.size[0], self.size[1]))
        for c, l in zip(self.coords, self.labels):
            label_map[c] = l

        # Ensure all labels are correct
        vi = 0
        for k in range(10):
            for j in range(self.size[0]):
                for i in range(self.size[1]):
                    if self.mask_list[k][j, i]:

                        l1 = self.labels[vi]
                        if k > 0 and self.mask_list[k - 1][j, i]:
                            l2 = label_map[k - 1, j, i]
                            assert l2 == l1
                        if j > 0 and self.mask_list[k][j - 1, i]:
                            l2 = label_map[k, j - 1, i]
                            assert l2 == l1
                        if i > 0 and self.mask_list[k][j, i - 1]:
                            l2 = label_map[k, j, i - 1]
                            assert l2 == l1
                        vi += 1



"""
Unit testing for the export_mtz.py routines
"""


from __future__ import annotations

import itertools

from dials.util.batch_handling import calculate_batch_offsets


def range_to_set(ranges):
    """Convert a list of ranges to a set of numbers"""
    return set().union(*[set(range(l, h + 1)) for l, h in ranges])


def has_consecutive_ranges(ranges):
    """Check that a set of ranges is non-consecutive"""
    total_entries = sum(h - l + 1 + 1 for l, h in ranges)
    union = set().union(*[set(range(l, h + 2)) for l, h in ranges])
    return len(union) < total_entries


def offset_ranges(offsets, ranges):
    return [(l + off, h + off) for off, (l, h) in zip(offsets, ranges)]


class TestBatchRangeCalculations:
    "Test the calculation of non-overlapping batch ranges"

    class MockScan:
        def __init__(self, image_range):
            self._batch_offset = 0
            self._image_range = image_range

        def get_batch_offset(self):
            return self._batch_offset

        def set_batch_offset(self, batch_offset):
            self._batch_offset = batch_offset

        def get_image_range(self):
            return self._image_range

    class MockExperiment:
        def __init__(self, image_range, scan=True):
            assert len(image_range) == 2
            self.scaling_model = None
            if scan:
                self.scan = TestBatchRangeCalculations.MockScan(image_range)
            else:
                self.scan = []

    def _run_ranges(self, ranges):
        """Convenience method to run the routine with a minimal experiment, and return the result as ranges of batch number"""
        input_data = [self.MockExperiment(x) for x in ranges]
        return offset_ranges(calculate_batch_offsets(input_data), ranges)

    def _run_ranges_to_set(self, ranges):
        """Runs a list of ranges and returns a set of individual batch numbers"""
        return range_to_set(self._run_ranges(ranges))

    def test_calculate_batch_ranges(self):
        assert self._run_ranges([(1, 1)]) == [(1, 1)]

        # Zero is shifted
        assert all(
            x > 0 for x in self._run_ranges_to_set([(0, 0)])
        ), "Should be no zeroth/negative batch"

        assert not set(self._run_ranges([(1, 1), (1, 1)])) == {
            (1, 1)
        }, "Overlapping simple ranges"

        data_tests = [
            [(1, 1), (1, 1)],
            # while we decide behaviour, remove input
            #            [(1, 1), (8, 8), (9, 9)],
            [(23, 24), (70, 100), (1, 1), (1, 4), (1, 1)],
            [(0, 98)],
        ]
        for data in data_tests:
            print("Running ", data)
            print("  ", self._run_ranges(data))
            assert all(
                float(x).is_integer() for x in itertools.chain(*self._run_ranges(data))
            ), "Fractional epochs"
            assert all(
                isinstance(x, int) for x in itertools.chain(*self._run_ranges(data))
            ), "Not all true integers"
            assert all(
                x > 0 for x in self._run_ranges_to_set([(0, 0)])
            ), "Should be no zeroth/negative batch"
            assert not has_consecutive_ranges(self._run_ranges(data))

        exp1 = TestBatchRangeCalculations.MockExperiment((1, 1), scan=False)
        exp2 = TestBatchRangeCalculations.MockExperiment((1, 1), scan=False)
        offsets = calculate_batch_offsets([exp1, exp2])
        assert all(float(x).is_integer() for x in offsets)
        assert all(isinstance(x, int) for x in offsets)
        assert all(x > 0 for x in offsets)



"""
Test for the functions in dials.util.batch_handling.

These functions are designed to handle reflection/experiment lists that
include both sequence experiments and single image datasets, which do not
have a scan object.
"""

from __future__ import annotations

from dxtbx.model import Experiment, Scan

from dials.array_family import flex
from dials.util.batch_handling import (
    _next_epoch,
    assign_batches_to_reflections,
    calculate_batch_offsets,
    get_batch_ranges,
    get_image_ranges,
    set_batch_offsets,
)


def reflections_1():
    """Test reflection table with batch"""
    r = flex.reflection_table()
    r["batch"] = flex.int([1, 11, 21, 31, 41, 51, 61, 71, 81, 91])
    r.set_flags(flex.bool(10, False), r.flags.user_excluded_in_scaling)
    return r


def reflections_2():
    """Test reflection table with batch"""
    r = flex.reflection_table()
    r["batch"] = flex.int([201, 211, 221, 231, 241, 251, 261, 271, 281, 291])
    r.set_flags(flex.bool(10, False), r.flags.user_excluded_in_scaling)
    return r


def reflections_3():
    """Test reflection table with xyzobs.px.value"""
    r = flex.reflection_table()
    r["xyzobs.px.value"] = flex.vec3_double([(0, 0, 0.5), (0, 0, 1.5)])
    return r


def test_assign_batches_to_reflections():
    """Test for namesake function"""
    reflections = [reflections_3(), reflections_3()]
    reflections = assign_batches_to_reflections(reflections, batch_offsets=[0, 100])
    assert list(reflections[0]["batch"]) == [1, 2]
    assert list(reflections[1]["batch"]) == [101, 102]


def test_calculate_batch_offsets():
    """Test offset calculation. Offset is next number ending in 01 bigger than
    previous batch numbers which is not consecutive"""
    scan = Scan(image_range=[1, 200], oscillation=[0.0, 1.0])
    exp1 = Experiment(scan=scan)
    exp2 = Experiment()
    offsets = calculate_batch_offsets([exp1, exp2])
    assert offsets == [0, 301]


def test_set_batch_offsets():
    """Test for namesake function"""
    scan = Scan(image_range=[1, 200], oscillation=[0.0, 1.0])
    exp1 = Experiment(scan=scan)
    exp2 = Experiment()
    set_batch_offsets([exp2, exp1], [0, 101])
    assert exp1.scan.get_batch_offset() == 101


def test_get_batch_ranges():
    """Test for namesake function"""
    scan = Scan(image_range=[1, 200], oscillation=[0.0, 1.0])
    exp1 = Experiment(scan=scan)
    exp2 = Experiment(scan=scan)
    batch_offsets = [0, 300]
    experiments = [exp1, exp2]
    batch_ranges = get_batch_ranges(experiments, batch_offsets)
    assert batch_ranges == [(1, 200), (301, 500)]


def test_get_image_ranges():
    """Test for namesake function"""
    scan = Scan(image_range=[1, 200], oscillation=[0.0, 1.0])
    exp1 = Experiment(scan=scan)
    exp2 = Experiment()
    experiments = [exp1, exp2]
    image_ranges = get_image_ranges(experiments)
    assert image_ranges == [(1, 200), (0, 0)]


def test_next_epoch():
    """Test for namesake function"""
    assert _next_epoch(100) == 201
    assert _next_epoch(99) == 101
    assert _next_epoch(105) == 201



from __future__ import annotations

import errno
import os

import matplotlib

from libtbx.phil import parse

import dials.util
from dials.algorithms.refinement.rotation_decomposition import (
    solve_r3_rotation_for_angles_given_axes,
)

matplotlib.use("Agg")
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt

phil_scope = parse(
    """
  output {
    directory = .
      .type = str
      .help = "The directory to store the results"

    format = *png pdf
      .type = choice

    debug = False
      .help = "print tables of values that will be plotted"
      .type = bool
      .expert_level = 1
  }

  orientation_decomposition
    .help = "Options determining how the orientation matrix"
            "decomposition is done. The axes about which to decompose"
            "the matrix into three rotations are chosen here, as well"
            "as whether the rotations are relative to the reference"
            "orientation, taken from the static crystal model"
  {
    e1 = 1. 0. 0.
      .type = floats(size = 3)

    e2 = 0. 1. 0.
      .type = floats(size = 3)

    e3 = 0. 0. 1.
      .type = floats(size = 3)

    relative_to_static_orientation = True
      .type = bool
  }
"""
)

help_message = """

Generate plots of scan-varying models, including crystal orientation, unit cell
and beam centre, from the input refined.expt

Examples::

  dials.plot_scan_varying_model refined.expt
"""


def ensure_directory(path):
    """Make the directory if not already there."""
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


class Script:
    """Class to run script."""

    def __init__(self):
        """Setup the script."""
        from dials.util.options import ArgumentParser

        usage = "usage: dials.plot_scan_varying_model [options] refined.expt"
        self.parser = ArgumentParser(
            usage=usage,
            phil=phil_scope,
            read_experiments=True,
            check_format=False,
            epilog=help_message,
        )

    def run(self, args=None):
        """Run the script."""
        from scitbx import matrix

        from dials.util.options import flatten_experiments

        params, options = self.parser.parse_args(args)
        if len(params.input.experiments) == 0:
            self.parser.print_help()
            return
        experiments = flatten_experiments(params.input.experiments)

        # Determine output path
        self._directory = os.path.join(params.output.directory, "scan-varying_model")
        self._directory = os.path.abspath(self._directory)
        ensure_directory(self._directory)
        self._format = "." + params.output.format

        self._debug = params.output.debug

        # Decomposition axes
        self._e1 = params.orientation_decomposition.e1
        self._e2 = params.orientation_decomposition.e2
        self._e3 = params.orientation_decomposition.e3

        # cell plot
        dat = []
        for iexp, exp in enumerate(experiments):

            crystal = exp.crystal
            scan = exp.scan

            if crystal.num_scan_points == 0:
                print("Ignoring scan-static crystal")
                continue

            scan_pts = list(range(crystal.num_scan_points))
            cells = [crystal.get_unit_cell_at_scan_point(t) for t in scan_pts]
            cell_params = [e.parameters() for e in cells]
            a, b, c, aa, bb, cc = zip(*cell_params)
            start, stop = scan.get_array_range()
            phi = [scan.get_angle_from_array_index(t) for t in range(start, stop + 1)]
            vol = [e.volume() for e in cells]
            cell_dat = {
                "phi": phi,
                "a": a,
                "b": b,
                "c": c,
                "alpha": aa,
                "beta": bb,
                "gamma": cc,
                "volume": vol,
            }
            try:
                cell_esds = [
                    crystal.get_cell_parameter_sd_at_scan_point(t) for t in scan_pts
                ]
                sig_a, sig_b, sig_c, sig_aa, sig_bb, sig_cc = zip(*cell_esds)
                cell_dat["sig_a"] = sig_a
                cell_dat["sig_b"] = sig_b
                cell_dat["sig_c"] = sig_c
                cell_dat["sig_aa"] = sig_aa
                cell_dat["sig_bb"] = sig_bb
                cell_dat["sig_cc"] = sig_cc
            except RuntimeError:
                pass

            if self._debug:
                print(f"Crystal in Experiment {iexp}")
                print("Phi\ta\tb\tc\talpha\tbeta\tgamma\tVolume")
                msg = "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}"
                line_dat = zip(phi, a, b, c, aa, bb, cc, vol)
                for line in line_dat:
                    print(msg.format(*line))
            dat.append(cell_dat)
        if dat:
            self.plot_cell(dat)

        # orientation plot
        dat = []
        for iexp, exp in enumerate(experiments):

            crystal = exp.crystal
            scan = exp.scan

            if crystal.num_scan_points == 0:
                print("Ignoring scan-static crystal")
                continue

            scan_pts = list(range(crystal.num_scan_points))
            start, stop = scan.get_array_range()
            phi = [scan.get_angle_from_array_index(t) for t in range(start, stop + 1)]
            Umats = [matrix.sqr(crystal.get_U_at_scan_point(t)) for t in scan_pts]
            if params.orientation_decomposition.relative_to_static_orientation:
                # factor out static U
                Uinv = matrix.sqr(crystal.get_U()).inverse()
                Umats = [U * Uinv for U in Umats]
            # NB e3 and e1 definitions for the crystal are swapped compared
            # with those used inside the solve_r3_rotation_for_angles_given_axes
            # method
            angles = [
                solve_r3_rotation_for_angles_given_axes(
                    U, self._e3, self._e2, self._e1, deg=True
                )
                for U in Umats
            ]
            phi3, phi2, phi1 = zip(*angles)
            angle_dat = {"phi": phi, "phi3": phi3, "phi2": phi2, "phi1": phi1}
            if self._debug:
                print(f"Crystal in Experiment {iexp}")
                print("Image\tphi3\tphi2\tphi1")
                msg = "{0}\t{1}\t{2}\t{3}"
                line_dat = zip(phi, phi3, phi2, phi1)
                for line in line_dat:
                    print(msg.format(*line))
            dat.append(angle_dat)
        if dat:
            self.plot_orientation(dat)

        # beam centre plot
        dat = []
        for iexp, exp in enumerate(experiments):

            beam = exp.beam
            detector = exp.detector
            scan = exp.scan

            if beam.num_scan_points == 0:
                print("Ignoring scan-static beam")
                continue

            scan_pts = range(beam.num_scan_points)
            start, stop = scan.get_array_range()
            phi = [scan.get_angle_from_array_index(t) for t in range(start, stop + 1)]
            p = detector.get_panel_intersection(beam.get_s0())
            if p < 0:
                print("Beam does not intersect a panel")
                continue
            panel = detector[p]
            s0_scan_points = [
                beam.get_s0_at_scan_point(i) for i in range(beam.num_scan_points)
            ]
            bc_scan_points = [panel.get_beam_centre_px(s0) for s0 in s0_scan_points]
            bc_x, bc_y = zip(*bc_scan_points)
            dat.append({"phi": phi, "beam_centre_x": bc_x, "beam_centre_y": bc_y})
        if dat:
            self.plot_beam_centre(dat)

    def plot_cell(self, dat):
        plt.figure(figsize=(13, 10))
        gs = gridspec.GridSpec(4, 2, wspace=0.4, hspace=0.6)

        ax = plt.subplot(gs[0, 0])
        ax.ticklabel_format(useOffset=False)
        for cell in dat:
            if "sig_a" in cell:
                ax.errorbar(
                    cell["phi"][0::20], cell["a"][0::20], yerr=cell["sig_a"][0::20]
                )
            plt.plot(cell["phi"], cell["a"])
        plt.xlabel(r"rotation angle $\left(^\circ\right)$")
        plt.ylabel(r"length $\left(\AA\right)$")
        plt.title("a")

        ax = plt.subplot(gs[0, 1])
        ax.ticklabel_format(useOffset=False)
        ymin, ymax = 180.0, 0.0
        for cell in dat:
            if "sig_aa" in cell:
                ax.errorbar(
                    cell["phi"][0::20], cell["alpha"][0::20], yerr=cell["sig_aa"][0::20]
                )
            plt.plot(cell["phi"], cell["alpha"])
            # choose the widest y range
            ymin = min(ymin, min(cell["alpha"]) - 0.1)
            ymax = max(ymax, max(cell["alpha"]) + 0.1)
            plt.axis(ymin=ymin, ymax=ymax)
        plt.xlabel(r"rotation angle $\left(^\circ\right)$")
        plt.ylabel(r"angle $\left(^\circ\right)$")
        plt.title(r"$\alpha$")

        ax = plt.subplot(gs[1, 0])
        ax.ticklabel_format(useOffset=False)
        for cell in dat:
            if "sig_b" in cell:
                ax.errorbar(
                    cell["phi"][0::20], cell["b"][0::20], yerr=cell["sig_b"][0::20]
                )
            plt.plot(cell["phi"], cell["b"])
        plt.xlabel(r"rotation angle $\left(^\circ\right)$")
        plt.ylabel(r"length $\left(\AA\right)$")
        plt.title("b")

        ax = plt.subplot(gs[1, 1])
        ax.ticklabel_format(useOffset=False)
        ymin, ymax = 180.0, 0.0
        for cell in dat:
            if "sig_bb" in cell:
                ax.errorbar(
                    cell["phi"][0::20], cell["beta"][0::20], yerr=cell["sig_bb"][0::20]
                )
            plt.plot(cell["phi"], cell["beta"])
            # choose the widest y range
            ymin = min(ymin, min(cell["beta"]) - 0.1)
            ymax = max(ymax, max(cell["beta"]) + 0.1)
            plt.axis(ymin=ymin, ymax=ymax)
        plt.xlabel(r"rotation angle $\left(^\circ\right)$")
        plt.ylabel(r"angle $\left(^\circ\right)$")
        plt.title(r"$\beta$")

        ax = plt.subplot(gs[2, 0])
        ax.ticklabel_format(useOffset=False)
        for cell in dat:
            if "sig_c" in cell:
                ax.errorbar(
                    cell["phi"][0::20], cell["c"][0::20], yerr=cell["sig_c"][0::20]
                )
            plt.plot(cell["phi"], cell["c"])
        plt.xlabel(r"rotation angle $\left(^\circ\right)$")
        plt.ylabel(r"length $\left(\AA\right)$")
        plt.title("c")

        ax = plt.subplot(gs[2, 1])
        ax.ticklabel_format(useOffset=False)
        ymin, ymax = 180.0, 0.0
        for cell in dat:
            if "sig_cc" in cell:
                ax.errorbar(
                    cell["phi"][0::20], cell["gamma"][0::20], yerr=cell["sig_cc"][0::20]
                )
            plt.plot(cell["phi"], cell["gamma"])
            # choose the widest y range
            ymin = min(ymin, min(cell["gamma"]) - 0.1)
            ymax = max(ymax, max(cell["gamma"]) + 0.1)
            plt.axis(ymin=ymin, ymax=ymax)
        plt.xlabel(r"rotation angle $\left(^\circ\right)$")
        plt.ylabel(r"angle $\left(^\circ\right)$")
        plt.title(r"$\gamma$")

        ax = plt.subplot2grid((4, 2), (3, 0), colspan=2)
        ax.ticklabel_format(useOffset=False)
        for cell in dat:
            plt.plot(cell["phi"], cell["volume"])
        plt.xlabel(r"rotation angle $\left(^\circ\right)$")
        plt.ylabel(r"volume $\left(\AA^3\right)$")
        plt.title("Cell volume")

        basename = os.path.join(self._directory, "unit_cell")
        fullname = basename + self._format
        print(f"Saving unit cell plot to {fullname}")
        plt.savefig(fullname)

    def plot_orientation(self, dat):
        plt.figure(figsize=(13, 10))
        gs = gridspec.GridSpec(3, 1, wspace=0.4, hspace=0.6)

        ax = plt.subplot(gs[0, 0])
        ax.ticklabel_format(useOffset=False)
        for ori in dat:
            plt.plot(ori["phi"], ori["phi1"])
        plt.xlabel(r"rotation angle $\left(^\circ\right)$")
        plt.ylabel(r"angle $\left(^\circ\right)$")
        plt.title(r"$\phi_1$")

        ax = plt.subplot(gs[1, 0])
        ax.ticklabel_format(useOffset=False)
        for ori in dat:
            plt.plot(ori["phi"], ori["phi2"])
        plt.xlabel(r"rotation angle $\left(^\circ\right)$")
        plt.ylabel(r"angle $\left(^\circ\right)$")
        plt.title(r"$\phi_2$")

        ax = plt.subplot(gs[2, 0])
        ax.ticklabel_format(useOffset=False)
        for ori in dat:
            plt.plot(ori["phi"], ori["phi3"])
        plt.xlabel(r"rotation angle $\left(^\circ\right)$")
        plt.ylabel(r"angle $\left(^\circ\right)$")
        plt.title(r"$\phi_3$")

        basename = os.path.join(self._directory, "orientation")
        fullname = basename + self._format
        print(f"Saving orientation plot to {fullname}")
        plt.savefig(fullname)

    def plot_beam_centre(self, dat):
        plt.figure(figsize=(13, 10))
        gs = gridspec.GridSpec(2, 1, wspace=0.4, hspace=0.6)

        ax = plt.subplot(gs[0, 0])
        ax.ticklabel_format(useOffset=False)
        ymin, ymax = 0.0, 0.0
        for bc in dat:
            plt.plot(bc["phi"], bc["beam_centre_x"])
            ymin = max(ymin, min(bc["beam_centre_x"]) - 0.1)
            ymax = max(ymax, max(bc["beam_centre_x"]) + 0.1)
            plt.axis(ymin=ymin, ymax=ymax)
        plt.xlabel(r"rotation angle $\left(^\circ\right)$")
        plt.ylabel(r"X (pixels)")
        plt.title(r"Beam centre X (pixels)")

        ax = plt.subplot(gs[1, 0])
        ax.ticklabel_format(useOffset=False)
        ymin, ymax = 0.0, 0.0
        for bc in dat:
            plt.plot(bc["phi"], bc["beam_centre_y"])
            ymin = max(ymin, min(bc["beam_centre_y"]) - 0.1)
            ymax = max(ymax, max(bc["beam_centre_y"]) + 0.1)
            plt.axis(ymin=ymin, ymax=ymax)
        plt.xlabel(r"rotation angle $\left(^\circ\right)$")
        plt.ylabel(r"Y (pixels)")
        plt.title(r"Beam centre Y (pixels)")

        basename = os.path.join(self._directory, "beam_centre")
        fullname = basename + self._format
        print(f"Saving beam centre plot to {fullname}")
        plt.savefig(fullname)


@dials.util.show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



from __future__ import annotations

import os
import pickle
import shutil

import procrunner
import pytest

from dxtbx.model.experiment_list import ExperimentListFactory

import dials.command_line.find_spots
from dials.array_family import flex


def _check_expected_results(reflections):
    """Check expected results for standard processing."""
    assert len(reflections) in range(653, 655)
    refl = reflections[0]
    assert refl["intensity.sum.value"] == pytest.approx(42)
    assert refl["bbox"] == pytest.approx((1398, 1400, 513, 515, 0, 1))
    assert refl["xyzobs.px.value"] == pytest.approx(
        (1399.1190476190477, 514.2142857142857, 0.5)
    )
    assert "shoebox" in reflections


def test_find_spots_from_images(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.find_spots",
            "nproc=1",
            "output.reflections=spotfinder.refl",
            "output.shoeboxes=True",
            "algorithm=dispersion",
        ]
        + list(dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "spotfinder.refl").is_file()

    reflections = flex.reflection_table.from_file(tmp_path / "spotfinder.refl")
    _check_expected_results(reflections)
    # No identifiers set if just running on images and not outputting experiments
    assert not reflections.experiment_identifiers().values(), list(
        reflections.experiment_identifiers().values()
    )


def test_find_spots_from_images_override_maximum(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.find_spots",
            "nproc=1",
            "maximum_trusted_value=100",
            "output.reflections=spotfinder.refl",
            "output.shoeboxes=True",
            "algorithm=dispersion",
        ]
        + list(dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "spotfinder.refl").is_file()

    reflections = flex.reflection_table.from_file(tmp_path / "spotfinder.refl")
    sbox = reflections["shoebox"]
    for s in sbox:
        assert flex.max(s.data) <= 100


def test_find_spots_from_zero_indexed_cbf(dials_data, tmp_path):
    one_indexed_cbf = (
        dials_data("centroid_test_data", pathlib=True) / "centroid_0001.cbf"
    )
    zero_indexed_cbf = tmp_path / "centroid_0000.cbf"
    shutil.copy(one_indexed_cbf, zero_indexed_cbf)

    result = procrunner.run(
        ["dials.find_spots", "nproc=1", zero_indexed_cbf], working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "strong.refl").is_file()
    assert b"Saved 0 reflections to" not in result.stdout, "No spots found on 0000.cbf"


def test_find_spots_from_images_output_experiments(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.find_spots",
            "nproc=1",
            "output.reflections=spotfinder.refl",
            "output.shoeboxes=True",
            "algorithm=dispersion",
            "output.experiments=spotfinder.expt",
        ]
        + list(dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "spotfinder.refl").is_file()

    reflections = flex.reflection_table.from_file(tmp_path / "spotfinder.refl")
    _check_expected_results(reflections)
    # Identifiers set if experiments are output
    assert reflections.experiment_identifiers().values(), list(
        reflections.experiment_identifiers().values()
    )


def test_find_spots_from_imported_experiments(dials_data, tmp_path):
    """First run import to generate an imported.expt and use this."""
    _ = procrunner.run(
        ["dials.import"]
        + list(dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")),
        working_directory=tmp_path,
    )

    result = procrunner.run(
        [
            "dials.find_spots",
            "nproc=1",
            tmp_path / "imported.expt",
            "output.reflections=spotfinder.refl",
            "output.shoeboxes=True",
            "algorithm=dispersion",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "spotfinder.refl").is_file()

    reflections = flex.reflection_table.from_file(tmp_path / "spotfinder.refl")
    _check_expected_results(reflections)
    # Identifiers set if just running on images
    assert len(reflections.experiment_identifiers().values()) == 1, list(
        reflections.experiment_identifiers().values()
    )


def test_find_spots_from_imported_as_grid(dials_data, tmp_path):
    """First run import to generate an imported.expt and use this."""
    _ = procrunner.run(
        ["dials.import", "oscillation=0,0"]
        + list(dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")),
        working_directory=tmp_path,
    )

    result = procrunner.run(
        [
            "dials.find_spots",
            "nproc=1",
            tmp_path / "imported.expt",
            "output.reflections=spotfinder.refl",
            "output.shoeboxes=True",
            "algorithm=dispersion",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "spotfinder.refl").is_file()

    reflections = flex.reflection_table.from_file(tmp_path / "spotfinder.refl")
    assert len(set(reflections["id"])) == 9, len(set(reflections["id"]))


def test_find_spots_with_resolution_filter(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.find_spots",
            "nproc=1",
            "output.reflections=spotfinder.refl",
            "output.shoeboxes=False",
            "algorithm=dispersion",
            "filter.d_min=2",
            "filter.d_max=15",
        ]
        + list(dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "spotfinder.refl").is_file()

    reflections = flex.reflection_table.from_file(tmp_path / "spotfinder.refl")
    assert len(reflections) in range(467, 469)
    assert "shoebox" not in reflections


def test_find_spots_with_hot_mask(dials_data, tmp_path):
    # now write a hot mask
    result = procrunner.run(
        [
            "dials.find_spots",
            "nproc=1",
            "write_hot_mask=True",
            "output.reflections=spotfinder.refl",
            "algorithm=dispersion",
            "output.shoeboxes=False",
        ]
        + list(dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "spotfinder.refl").is_file()
    assert (tmp_path / "hot_mask_0.pickle").is_file()

    reflections = flex.reflection_table.from_file(tmp_path / "spotfinder.refl")
    assert len(reflections) in range(653, 655)
    assert "shoebox" not in reflections

    with (tmp_path / "hot_mask_0.pickle").open("rb") as f:
        mask = pickle.load(f)
    assert len(mask) == 1
    assert mask[0].count(False) == 12


def test_find_spots_with_hot_mask_with_prefix(dials_data, tmp_path):
    # now write a hot mask
    result = procrunner.run(
        [
            "dials.find_spots",
            "nproc=1",
            "write_hot_mask=True",
            "hot_mask_prefix=my_hot_mask",
            "output.reflections=spotfinder.refl",
            "output.shoeboxes=False",
            "algorithm=dispersion",
        ]
        + list(dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "spotfinder.refl").is_file()
    assert (tmp_path / "my_hot_mask_0.pickle").is_file()

    reflections = flex.reflection_table.from_file(tmp_path / "spotfinder.refl")
    assert len(reflections) in range(653, 655)
    assert "shoebox" not in reflections
    with (tmp_path / "my_hot_mask_0.pickle").open("rb") as f:
        mask = pickle.load(f)
    assert len(mask) == 1
    assert mask[0].count(False) == 12


def test_find_spots_with_generous_parameters(dials_data, tmp_path):
    # now with more generous parameters
    result = procrunner.run(
        [
            "dials.find_spots",
            "nproc=1",
            "min_spot_size=3",
            "max_separation=3",
            "output.reflections=spotfinder.refl",
            "algorithm=dispersion",
        ]
        + list(dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "spotfinder.refl").is_file()

    reflections = flex.reflection_table.from_file(tmp_path / "spotfinder.refl")
    assert len(reflections) in range(678, 680)


def test_find_spots_with_user_defined_mask(dials_data, tmp_path):
    # Now with a user defined mask
    mask_pickle = dials_data("centroid_test_data", pathlib=True) / "mask.pickle"
    result = procrunner.run(
        [
            "dials.find_spots",
            "nproc=1",
            "output.reflections=spotfinder.refl",
            "output.shoeboxes=True",
            "algorithm=dispersion",
            f"lookup.mask={mask_pickle}",
        ]
        + list(dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "spotfinder.refl").is_file()

    reflections = flex.reflection_table.from_file(tmp_path / "spotfinder.refl")
    experiments = ExperimentListFactory.from_json_file(
        dials_data("centroid_test_data", pathlib=True) / "experiments.json"
    )
    assert len(experiments) == 1
    imageset = experiments.imagesets()[0]
    detector = imageset.get_detector()
    beam = imageset.get_beam()
    for x, y, z in reflections["xyzobs.px.value"]:
        d = detector[0].get_resolution_at_pixel(beam.get_s0(), (x, y))
        assert d >= 3


def test_find_spots_with_user_defined_region(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.find_spots",
            "nproc=1",
            "output.reflections=spotfinder.refl",
            "output.shoeboxes=True",
            "region_of_interest=800,1200,800,1200",
        ]
        + list(dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "spotfinder.refl").is_file()

    reflections = flex.reflection_table.from_file(tmp_path / "spotfinder.refl")
    x, y, z = reflections["xyzobs.px.value"].parts()
    assert x.all_ge(800)
    assert y.all_ge(800)
    assert x.all_lt(1200)
    assert y.all_lt(1200)


def test_find_spots_with_xfel_stills(dials_regression, tmp_path):
    # now with XFEL stills
    result = procrunner.run(
        [
            "dials.find_spots",
            "nproc=1",
            os.path.join(
                dials_regression,
                "spotfinding_test_data",
                "idx-s00-20131106040302615.cbf",
            ),
            "output.reflections=spotfinder.refl",
            "algorithm=dispersion",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "spotfinder.refl").is_file()

    reflections = flex.reflection_table.from_file(tmp_path / "spotfinder.refl")
    assert len(reflections) == 2643


def test_find_spots_with_per_image_statistics(dials_data, tmp_path):
    result = procrunner.run(
        ["dials.find_spots", "nproc=1", "per_image_statistics=True"]
        + list(dials_data("centroid_test_data", pathlib=True).glob("centroid*.cbf")),
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "strong.refl").is_file()
    assert b"Number of centroids per image for imageset 0:" in result.stdout
    assert (
        b"|   image |   #spots |   #spots_no_ice |   total_intensity |" in result.stdout
    )


@pytest.mark.parametrize(
    "blur,expected_nref", [("None", 559), ("narrow", 721), ("wide", 739)]
)
def test_find_spots_radial_profile(dials_data, blur, expected_nref, run_in_tmp_path):
    reflections = dials.command_line.find_spots.run(
        [
            "nproc=1",
            "threshold.algorithm=radial_profile",
            f"blur={blur}",
        ]
        + [
            os.fspath(f)
            for f in dials_data("centroid_test_data", pathlib=True).glob(
                "centroid*.cbf"
            )
        ],
        return_results=True,
    )
    assert len(reflections) == expected_nref



"""
Test the observer module.
"""

from __future__ import annotations

from dials.util.observer import Observer, Subject, singleton


def test_singleton_decorator():
    """Test the singleton class decorator."""

    class TestClass:
        """Simplest class."""

        pass

    singleton_class = singleton(TestClass)

    c1 = singleton_class()
    c2 = singleton_class()

    assert c1 is c2

    c1 = TestClass()
    c2 = TestClass()

    assert c1 is not c2


def test_Observer():
    """Test for expected properties of the general observer."""

    observer = Observer()
    assert hasattr(observer, "data")
    observer.update([])


def test_Subject():
    """Test the subject class."""

    # First register events
    subject = Subject(events=["event1", "event2"])
    assert "event1" in subject.observers
    assert "event2" in subject.observers

    # Test the get_observers and register_observer and unregister_observer method
    observer = Observer()
    assert subject.get_observers("event1") == {}
    assert subject.get_observers("event2") == {}
    subject.register_observer("event1", observer)
    assert subject.get_observers("event1") == {observer: observer.update}
    assert subject.get_observers("event2") == {}
    subject.unregister_observer("event1", observer)
    assert subject.get_observers("event1") == {}
    assert subject.get_observers("event2") == {}

    # Try registering the observer with a different callback
    class MyObserver(Observer):
        """Test observer implementation class."""

        def __init__(self):
            self.count = 0

        def mycallback(self, _):
            """Custom callback method."""
            self.count += 1

    observer = MyObserver()
    subject = Subject(events=["event1", "event2"])
    subject.register_observer("event1", observer, callback="mycallback")
    assert subject.get_observers("event1") == {observer: observer.mycallback}
    assert subject.get_observers("event2") == {}

    # Now test call to notify.
    subject.notify("event1")
    assert observer.count == 1
    subject.notify("event2")
    assert observer.count == 1

    # Now test the notify decorator
    class MySubject(Subject):
        """Test subject implementation class."""

        def __init__(self):
            super().__init__(events=["event1"])

        @Subject.notify_event("event1")
        def test_method(self):
            """Test method for warpping."""
            return "test_return_value"

    mysubject = MySubject()
    myobserver = MyObserver()
    mysubject.register_observer("event1", myobserver, callback="mycallback")
    x = mysubject.test_method()
    assert myobserver.count == 1
    assert x == "test_return_value"



from __future__ import annotations

import logging

import iotbx.phil

import dials.util
from dials.array_family import flex
from dials.util import log
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.version import dials_version

help_message = """
Small modifications to an indexed spot list to allow it to be treated as if
it were integrated, for rapid feedback analysis. Do not use output data for
structure solution or refinement.
"""

phil_scope = iotbx.phil.parse(
    """
output {
  reflections = pseudo_integrated.refl
    .type = path
  log = "dials.indexed_as_integrated.log"
    .type = path
}
"""
)


logger = logging.getLogger("dials.command_line.indexed_as_integrated")


def indexed_as_integrated(reflections, params, experiments):
    """Small modifications to an indexed spot list to allow it to be
    treated as if it were integrated, for rapid feedback analysis."""

    # filter only indexed reflections & assign the summation integrated flag

    sel = reflections.get_flags(reflections.flags.indexed)
    reflections = reflections.select(sel)
    all = flex.bool(reflections.size(), True)
    reflections.set_flags(all, reflections.flags.integrated_sum)

    logger.info(f"\nSaved {sel.count(True)} of {sel.size()} reflections\n")

    # add resolution to reflections

    if "imageset_id" not in reflections:
        reflections["imageset_id"] = reflections["id"]

    reflections.centroid_px_to_mm(experiments)
    reflections.map_centroids_to_reciprocal_space(experiments)
    reflections["d"] = 1 / reflections["rlp"].norms()

    return reflections


@dials.util.show_mail_handle_errors()
def run(args=None):
    usage = "dials.indexed_as_integrated [options] indexed.refl indexed.expt"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    params, _ = parser.parse_args(args, show_diff_phil=True)

    log.config(logfile=params.output.log)
    logger.info(
        dials_version()
        + "\n\n----------------------------------------------------------\n"
        "Do not use the output for structure solution or refinement"
        "\n----------------------------------------------------------"
    )

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    if len(reflections) != 1 or len(experiments) != 1:
        parser.print_help()
        return

    reflections = reflections[0]

    pseudo_integrated = indexed_as_integrated(reflections, params, experiments)
    pseudo_integrated.as_file(params.output.reflections)

    logger.info(f"\nSaved pseudo-integrated data to: {params.output.reflections}\n")


if __name__ == "__main__":
    run()



from __future__ import annotations

import procrunner


def test_merge_cbf(dials_data, tmp_path):
    data_dir = dials_data("centroid_test_data", pathlib=True)

    g = sorted(data_dir.glob("*.cbf"))
    assert len(g) == 9

    cmd = ["dials.merge_cbf", "merge_n_images=3"] + g
    result = procrunner.run(cmd, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    g = sorted(tmp_path.glob("sum_*.cbf"))
    assert len(g) == 3

    # test alternate mode of accessing image data
    cmd += ["image_prefix=sum2_", "get_raw_data_from_imageset=false"]
    result = procrunner.run(cmd, working_directory=tmp_path)
    assert not result.returncode and not result.stderr

    g2 = sorted(tmp_path.glob("sum2_*.cbf"))
    assert len(g2) == 3

    # check summed images are the same in either case
    for f1, f2 in zip(g, g2):
        print("Testing", f1, f2)
        assert f1.read_bytes() == f2.read_bytes()



from __future__ import annotations

import math

from dxtbx.model.experiment_list import ExperimentList, ExperimentListFactory

from dials.algorithms.profile_model.gaussian_rs import Model, PartialityCalculator3D
from dials.array_family import flex


def test(dials_data):
    exlist = ExperimentListFactory.from_json_file(
        dials_data("centroid_test_data", pathlib=True) / "fake_long_experiments.json"
    )

    assert len(exlist) == 1
    experiment = exlist[0]

    # Set the delta_divergence/mosaicity
    n_sigma = 5
    sigma_b = 0.060 * math.pi / 180
    sigma_m = 0.154 * math.pi / 180

    profile_model = Model(None, n_sigma, sigma_b, sigma_m)
    experiment.profile = profile_model
    experiments = ExperimentList()
    experiments.append(experiment)

    calculator = PartialityCalculator3D(
        experiment.beam, experiment.goniometer, experiment.scan, sigma_m
    )

    predicted = flex.reflection_table.from_predictions_multi(experiments)
    predicted["bbox"] = predicted.compute_bbox(experiments)

    # Remove any touching edges of scan to get only fully recorded
    x0, x1, y0, y1, z0, z1 = predicted["bbox"].parts()
    predicted = predicted.select((z0 > 0) & (z1 < 100))
    assert len(predicted) > 0

    # Compute partiality
    partiality = calculator(
        predicted["s1"], predicted["xyzcal.px"].parts()[2], predicted["bbox"]
    )

    # Should have all fully recorded
    assert len(partiality) == len(predicted)
    three_sigma = 0.5 * (
        math.erf(3.0 / math.sqrt(2.0)) - math.erf(-3.0 / math.sqrt(2.0))
    )
    assert partiality.all_gt(three_sigma)

    # Trim bounding boxes
    x0, x1, y0, y1, z0, z1 = predicted["bbox"].parts()
    z0 = z0 + 1
    z1 = z1 - 1
    predicted["bbox"] = flex.int6(x0, x1, y0, y1, z0, z1)
    predicted = predicted.select(z1 > z0)
    assert len(predicted) > 0

    # Compute partiality
    partiality = calculator(
        predicted["s1"], predicted["xyzcal.px"].parts()[2], predicted["bbox"]
    )

    # Should have all partials
    assert len(partiality) == len(predicted)
    assert partiality.all_le(1.0) and partiality.all_gt(0)
